/*
 **************************************************************************
 *
 * Boot-ROM-Code to load an operating system across a TCP/IP network.
 *
 * Module:  bootp.c
 * Purpose: Get information for client with BOOTP protocol
 * Entries: bootp
 *
 **************************************************************************
 *
 * Copyright (C) 1995 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "../headers/net.h"
#include "../headers/boot.h"
#include "../headers/packet.h"
#include "../headers/romlib.h"
#include "../headers/general.h"
#include "./bootpriv.h"
#include "./bootp.h"
#include "./tftp.h"



/*
 **************************************************************************
 * 
 * Global variables:
 */
static struct bootp *boot_rec = NULL;	/* bootp record			*/
static unsigned short boot_secs = 0;	/* number of seconds since boot	*/
static unsigned long boot_xid;		/* bootp transaction ID		*/
static unsigned short boot_try;		/* number of tries		*/



/*
 **************************************************************************
 * 
 * Return information from the vendor area of the BOOTP record. Note
 * that this will work only with RFC1048-compliant messages.
 *
 */
static unsigned char *get_vend(int id)
{
  int nop_count = 0;
  unsigned char *cp = boot_rec->bp_vend;

  /* Check if vendor area is RFC1048-compliant */
  if (memcmp(cp, VM_RFC1048, VM_SIZE))
	return(NULL);

  /* Now scan the vendor information field. For VEND_MAXNOP see bootp.h */
  cp += VM_SIZE;
  while (*cp != VEND_END && nop_count < VEND_MAXNOP) {
	if (*cp == VEND_NOP) {
		nop_count++;
		cp++;
		continue;
	}
	if (*cp == id)
		return(cp);
	cp += *(cp + 1) + 2;
  }
  if (*cp == VEND_END && id == VEND_END)
	return(cp);
  return(NULL);
}



/*
 **************************************************************************
 * 
 * Send a BOOTP request via UDP.
 *
 */
static void bootp_send(void)
{
  /* Assemble the BOOTP request */
  memset(boot_rec, 0, sizeof(struct bootp));
  memcpy(&(boot_rec->bp_vend), VM_RFC1048, VM_SIZE);
  get_hw_addr((char *)&(boot_rec->bp_chaddr), sizeof(boot_rec->bp_chaddr));
  boot_rec->bp_op = BOOTP_REQUEST;
  boot_rec->bp_hwtype = BOOTP_ETHER;
  boot_rec->bp_hlen = ETH_ALEN;
  boot_rec->bp_xid = boot_xid;
  boot_rec->bp_secs = htons(boot_secs);
  boot_rec->bp_ciaddr = htonl(IP_ANY);
  boot_rec->bp_siaddr = htonl(IP_BROADCAST);

  /* Send the bootp record as a broadcast message to all servers */
  udp_write((char *)boot_rec, sizeof(struct bootp));
}



/*
 **************************************************************************
 * 
 * Get a BOOTP record from the server.
 *
 */
static int bootp_get(void)
{
  char errch;
  int mask, timeout, i;

  /*
   * First setup the BOOTP transaction ID. This is unique for each
   * individual BOOTP request.
   */
  boot_xid = get_ticks() + random();

  /*
   * Setup the IP interface information. Our own address has to be
   * IP_ANY, so that the IP layer will not discard the incoming packets.
   * Then open a UDP socket.
   */
  if_config(IP_ANY, IP_CLASS_A, IP_BROADCAST);
  udp_open(IP_BROADCAST, BOOTP_C_PORT, BOOTP_S_PORT);

  /*
   * Now loop until receiving a reply from a server. The retry time is
   * computed as suggested in RFC951.
   */
  mask = 0x01;
  timeout = 1;
  boot_try = 0;
  printf("Sending request: ");
  while(++boot_try < BOOTP_RETRY) {
	bootp_send();
	i = udp_read((char *)boot_rec, sizeof(struct bootp), timeout*18);
	errch = '\0';
	if (i == 0)
		errch = '.';
	else if (i != sizeof(struct bootp))
		errch = '*';
	else if (boot_rec->bp_xid != boot_xid)
		errch = '?';
	else if (boot_rec->bp_op != BOOTP_REPLY)
		errch = 'O';
	if (errch == '\0') break;
	if (mask < 0x3f)
		mask = (mask << 1) | 1;
	if ((timeout = mask & random()) < 1)
		timeout = 1;
	printf("%c", errch);
  }

  if (boot_try >= BOOTP_RETRY) {
	printf("\nBOOTP: No reply\n");
	return(FALSE);
  }

  /* Check if the BOOTP record is valid */
  if (get_vend(VEND_END) == NULL) {
	printf("\nBOOTP: Invalid record\n");
	return(FALSE);
  }

  printf("ok\n");
  return(TRUE);
}



#ifndef NOBPEXT
/*
 **************************************************************************
 * 
 * Get vendor extensions file with TFTP.
 */
static int get_ext(unsigned char *ftag)
{
  int size, i, fnamlen;
  char *inbuf, *fname;
  unsigned char *cp;
  struct bootp *bp, *temp;

  printf("BOOTP: Loading ");
  for (i = 0, cp = ftag + 2; *cp && i < *(ftag + 1); cp++, i++)
	printf("%c", *cp);
  printf(": ");

  /*
   * Allocate memory for a new BOOTP record. We don't accept extensions
   * of more than 2048 bytes, which is 4 TFTP blocks, and which should
   * be enough. Then copy the old record into the beginning of the
   * new one, and determine the beginning of the extension vendor area.
   */
  size = MAX_EXT_BLOCKS * SEGSIZE + sizeof(struct bootp) + 2;
  if ((bp = (struct bootp *)malloc(size)) == NULL) {
	printf("No memory\n");
	return(FALSE);
  }
  memcpy(bp, boot_rec, sizeof(struct bootp));
  cp = (unsigned char *)bp - (unsigned char *)boot_rec + get_vend(VEND_END);

  /* Open the TFTP connection */
  fname = (char *)(ftag + 2);
  fnamlen = *(ftag + 1);
  if ((inbuf = tftp_open(ntohl(boot_rec->bp_siaddr), fname, fnamlen)) == NULL)
	return(FALSE);

  /*
   * Read all blocks from vendor extension file. The file has the RFC
   * vendor ID at the beginning, so we have to take special care with
   * the first block.
   */
  for (i = 0; i < MAX_EXT_BLOCKS; i++) {
	if ((size = tftp_get()) < 0)
		return(FALSE);
	if (i == 0) {
		if (size < (VM_SIZE + 1) ||
		    memcmp(inbuf, VM_RFC1048, VM_SIZE)) {
			printf("Invalid file\n");
			return(FALSE);
		}
		memcpy(cp, (unsigned char *)inbuf + VM_SIZE, size - VM_SIZE);
		cp += size - VM_SIZE;
	} else {
		if (size == 0) break;
		memcpy(cp, (unsigned char *)inbuf, size);
		cp += size;
	}
	if (size < SEGSIZE)
		break;
  }

  /* If the file is too large, we are probably missing the end identifier */
  if (i == MAX_EXT_BLOCKS) {
	printf("File too large\n");
	return(FALSE);
  }

  /* Make the new BOOTP record active */
  temp = boot_rec;
  boot_rec = bp;
  free(temp);

  /* Finally check if we have an end identifier at all */
  if (get_vend(VEND_END) == NULL) {
	printf("Invalid file\n");
	return(FALSE);
  }
  printf("ok\n");
  return(TRUE);
}
#endif



#ifdef BPDEBUG
/*
 **************************************************************************
 * 
 * Print the complete BOOTP record
 */
static void print_rec(void)
{
  unsigned char *cp, *end;
  int i, c, adr;

  adr = 0;
  cp = (unsigned char *)boot_rec;
  if ((end = get_vend(VEND_END)) == NULL)
	end = cp + sizeof(struct bootp);
  printf("\n\n");
  while(cp <= end) {
	if (adr < 16) printf("00%x  ", adr);
	else if (adr < 256) printf("0%x  ", adr);
	else printf("%x  ", adr);

	for (i = 0; i < 16 && cp <= end; i++, cp++) {
		c = *cp & 0xff;
		if (c < 16) printf("0%x ", c);
		else printf("%x ", c);
	}
	printf("\n");
	adr += 16;
  }
  (void)getkey();
  printf("\n\n");
}
#endif



/*
 **************************************************************************
 * 
 * Handle BOOTP protocol.
 *
 */
struct bootp *bootp(void)
{
  unsigned char *cp;
  t_ipaddr netmask;
  t_ipaddr gateway;

  printf("BOOTP: ");
  if (boot_rec == NULL && (boot_rec = malloc(sizeof(struct bootp))) == NULL) {
	printf("No memory\n");
	return(NULL);
  }

  /* Get BOOTP record from server */
  if (!bootp_get())
	return(NULL);

  /* Setup the network interface with the values received from BOOTP server */
  netmask = IP_ANY;
  if ((cp = get_vend(VEND_SUBNET)) != NULL)
	netmask = ntohl(*((t_ipaddr *)(cp + 2)));
  if_config(ntohl(boot_rec->bp_yiaddr), netmask, IP_BROADCAST);

  /* Also setup the routing table so that we can reach the server. */
  if ((cp = get_vend(VEND_ROUTER)) != NULL) {
	gateway = ntohl(*((t_ipaddr *)(cp + 2)));
	rt_add(gateway, IP_ANY, netmask);
	rt_add(ntohl(boot_rec->bp_siaddr), gateway, IP_ANY);
	rt_add(IP_DEFAULT, gateway, IP_ANY);
  } else {
	rt_add(ntohl(boot_rec->bp_siaddr), IP_ANY, netmask);
	rt_add(IP_DEFAULT, IP_ANY, netmask);
  }

#ifndef NOBPEXT
  /* Next try to get the extensions file if there is any specified */
  if ((cp = get_vend(VEND_EXTFILE)) != NULL && !get_ext(cp))
	return(NULL);
#endif

#ifdef BPDEBUG
  print_rec();
#endif

  return(boot_rec);
}

