;**************************************************************************
;*
;* Boot-ROM-Code to load an operating system across a TCP/IP network.
;*
;* Module:  utility.asm
;* Purpose: Various functions for the image loader routines, which are
;*          better written in assembler
;* Entries: _convmem, _extmem, _lmove, _far2long, _long2far, _exec_image
;*
;**************************************************************************
;*
;* Copyright (C) 1995 Gero Kuhlmann <gero@gkminix.han.de>
;*
;*  This program is free software; you can redistribute it and/or modify
;*  it under the terms of the GNU General Public License as published by
;*  the Free Software Foundation; either version 2 of the License, or
;*  any later version.
;*
;*  This program is distributed in the hope that it will be useful,
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;*  GNU General Public License for more details.
;*
;*  You should have received a copy of the GNU General Public License
;*  along with this program; if not, write to the Free Software
;*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;*


;
;**************************************************************************
;
; Include assembler macros:
;
include ..\headers\asm\layout.inc
include ..\headers\asm\macros.inc
include ..\headers\asm\memory.inc


;
;**************************************************************************
;
; Layout of descriptors in GDT:
;
gdt		struc

gdt_length	dw	?		; length of segment
gdt_base	db	3 dup (?)	; base of segment
gdt_flags	db	?		; descriptor flags
gdt_res		dw	?		; reserved

gdt		ends

gdt_stdflag	equ	93h		; standard flags for GDT entries


;
;**************************************************************************
;
; Data segment
;
data_start

gdt_table	gdt	6 dup (<>)	; descriptor table for move routine

data_end


;
;**************************************************************************
;
; Start code segment.
;
text_start

	public	_convmem		; define entry points
	public	_extmem
	public	_lmove
	public	_long2far
	public	_far2long
	public	_exec_image

	extrn	_fatal:near


;
;**************************************************************************
;
; Get amount of conventional memory in bytes.
; Input:  none
; Output: memory amount
;
_convmem	proc	near

	int	12h			; just call the BIOS
	jmp	short extm1

_convmem	endp


;
;**************************************************************************
;
; Get amount of extended memory in bytes.
; Input:  none
; Output: memory amount
;
_extmem		proc	near

	mov	ah,88h
	int	15h			; just call the BIOS
extm1:	mov	dx,ax
	mov	cl,6			; multiply value with 1024 to get
	shr	dx,cl			; number of bytes
	mov	cl,10
	shl	ax,cl
	ret

_extmem		endp


;
;**************************************************************************
;
; Move a memory block into extended memory.
; Input:  1. arg  -  linear address to destination (long word)
;         2. arg  -  near pointer to source memory block
;         3. arg  -  length of memory block
; Output: non-zero if error
;
_lmove		proc	near

	push	bp
	mov	bp,sp
	getarg	bx,3
	getarg	ax,0
	mov	word ptr gdt_table[3 * type gdt].gdt_base[0],ax
	getarg	ax,1
	mov	gdt_table[3 * type gdt].gdt_base[2],al
	mov	gdt_table[3 * type gdt].gdt_flags,gdt_stdflag
	mov	gdt_table[3 * type gdt].gdt_length,bx
	mov	gdt_table[3 * type gdt].gdt_res,0
	mov	ax,ds
	mov	cl,4
	rol	ax,cl
	mov	cx,ax
	and	ax,0FFF0h		; convert segment and offset into
	and	cx,0000Fh		; a linear address
	getarg	dx,2
	add	ax,dx
	adc	cx,0
	mov	word ptr gdt_table[2 * type gdt].gdt_base[0],ax
	mov	gdt_table[2 * type gdt].gdt_base[2],cl
	mov	gdt_table[2 * type gdt].gdt_flags,gdt_stdflag
	mov	gdt_table[2 * type gdt].gdt_length,bx
	mov	gdt_table[2 * type gdt].gdt_res,0

	push	es
	push	si
	mov	cx,bx
	inc	cx			; convert byte count into word count
	shr	cx,1
	mov	ax,ds
	mov	es,ax
	mov	si,offset dgroup:gdt_table
	mov	ah,87h
	int	15h			; call BIOS to do the copy
	pop	si
	pop	es
	xor	al,al			; prepare return value
	xchg	al,ah
	pop	bp
	ret

_lmove		endp


;
;**************************************************************************
;
; Convert a far pointer to an unsigned long integer
; Input:  1. arg  -  far pointer
; Output: unsigned long integer
;
_far2long	proc	near

	push	bp
	mov	bp,sp
	getarg	dx,1
	mov	bx,dx
	mov	cl,12			; convert segment
	shr	dx,cl
	mov	cl,4
	shl	bx,cl
	getarg	ax,0
	add	ax,bx			; add offset to segment
	adc	dx,0
	pop	bp
	ret

_far2long	endp


;
;**************************************************************************
;
; Convert an unsigned long integer to a far pointer
; Input:  1. arg  -  unsigned long integer
; Output: far pointer
;
_long2far	proc	near

	push	bp
	mov	bp,sp
	getarg	dx,1
	mov	cl,12
	shl	dx,cl
	getarg	ax,0			; split the long integer into
	mov	bx,ax			; a segment and offset part
	mov	cl,4
	shr	bx,cl
	add	dx,bx
	and	ax,000Fh
	pop	bp
	ret

_long2far	endp


;
;**************************************************************************
;
; Execute the boot image
; Input: 1. arg  -  header mode
;        2. arg  -  size of header
;        3. arg  -  far pointer to execution point
;        4. arg  -  far pointer to header
;        5. arg  -  far pointer to parameter block
; Output: none
;
_exec_image	proc	near

	push	bp
	mov	bp,sp

; First reset the interrupt vector table. We don't need to save
; all segment and index registers since this routine will never
; return to the C code.

	cli				; make this copying atomic
	xor	ax,ax
	mov	es,ax
	mov	ax,OLDINTS
	mov	ds,ax
	xor	si,si
	xor	di,di
	mov	cx,VECTSIZE / 2
	rep	movsw
	sti

	getarg	ax,0
	cmp	ax,1			; check for mode 1
	jne	exec1

; Handle mode number 1. This means copying the header block to
; the execution point, and setting the stack just below there.
; Note that the image doesn't get anything on it's stack. It
; would be too complicated to check for overwriting the bootp
; block and/or stack with the header block.

	getarg	cx,1
	getarg	ax,3
	mov	es,ax			; get the pointer to the boot
	getarg	di,2			; area into ES:DI
	mov	bx,di
	getarg	ax,5
	mov	ds,ax			; put the pointer to the header
	getarg	si,4			; block into DS:SI
	rep	movsb			; now place the boot block into
	cli				; it's final place
	mov	ax,OLDINTS
	mov	sp,VECTSIZE		; use the old interrupt table as
	mov	ss,ax			; the new stack
	sti
	jmp	short exec9		; jump to the boot image

; Handle mode 2. This means just to assign a new stack, and then
; push the two parameters onto it before calling the new image.

exec1:	getarg	ax,3
	mov	es,ax			; get the execution pointer
	getarg	bx,2			; into ES:BX
	getarg	ax,5			; get the header pointer
	mov	ds,ax			; into DS:SI
	getarg	si,4
	getarg	cx,7			; get the parameter pointer
	getarg	di,6			; into CX:DI
	cli
	mov	ax,OLDINTS
	mov	sp,VECTSIZE		; use the old interrupt table as
	mov	ss,ax			; the new stack
	sti
	push	cx
	push	di			; push the parameter pointer
	push	ds
	push	si			; push the header pointer

; Jump to the boot image. The pointer is in ES:BX.

exec9:	push	cs
	mov	ax,offset cgroup:exech	; push the return pointer
	push	ax
	push	es			; jump to new image
	push	bx
	retf

exech:	jmp	_fatal

_exec_image	endp


;
;**************************************************************************
;
text_end

	end

