;**************************************************************************
;*
;* Boot-ROM-Code to load an operating system across a TCP/IP network.
;*
;* Module:  fileio.asm
;* Purpose: File I/O functions for DOS simulator
;* Entries: dos3C, dos3D, dos3E, dos3F, dos40
;*
;**************************************************************************
;*
;* Copyright (C) 1995 Gero Kuhlmann <gero@gkminix.han.de>
;*
;*  This program is free software; you can redistribute it and/or modify
;*  it under the terms of the GNU General Public License as published by
;*  the Free Software Foundation; either version 2 of the License, or
;*  any later version.
;*
;*  This program is distributed in the hope that it will be useful,
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;*  GNU General Public License for more details.
;*
;*  You should have received a copy of the GNU General Public License
;*  along with this program; if not, write to the Free Software
;*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;*


;
;**************************************************************************
;
; Include assembler macros:
;
include ..\headers\asm\layout.inc
include ..\headers\asm\macros.inc
include ..\headers\asm\memory.inc
include ..\headers\asm\minidos.inc
include .\dospriv.inc


;
;**************************************************************************
;
; Miscelleneous equates
;
CHR_BS		equ	08h		; Backspace character
CHR_CR		equ	0Dh		; Return character
CHR_LF		equ	0Ah		; Line-feed character
CHR_EOF		equ	1Ah		; End-of-file character


;
;**************************************************************************
;
; Data segment
;
data_start

		extrn	old_ds:word	; caller's DS register

data_end


;
;**************************************************************************
;
; Start code segment.
;
text_start

	extrn	devinit:near

	public	dos3C, dos3D, dos3E	; define entry points
	public	dos3F, dos40


;
;**************************************************************************
;
; File names recognized by the DOS simulator.
;
filetab	label	byte

	db	DEV_CON, 0FFh		; console device, no minor number
	db	'CON:',0		; name of console device

	db	DEV_AUX, 0FFh		; serial device, no minor number
	db	'AUX:',0		; name of serial device

	db	DEV_PRN, 0FFh		; printer device, no minor number
	db	'PRN:',0		; name of printer device

	db	DEV_AUX, 3		; serial device, minor number in pos 3
	db	'COM?:',0		; name of serial device

	db	DEV_PRN, 3		; parallel device, minor number in pos 3
	db	'LPT?:',0		; name of parallel device

	db	0FFh			; end of list


;
;**************************************************************************
;
; Create a new file. Since it's not possible to create a new device file
; (at least not with DOS :-)), this routine does nothing but calling the
; file opening routine.
; Input:  CX     -  file attribute
;         DS:DX  -  pointer to file name
; Output: AX     -  file handle or error code if carry flag set
;         Carry flag  -  set if error
; Registers changed: AX
;
dos3C		proc	near

	jmp	short dos3D

dos3C		endp


;
;**************************************************************************
;
; Open file. Since the DOS simulator cannot use files, it's only possible
; to open a device. This routine will just compute the file handle for
; that device. The device initialization is done with the read/write calls
; because the standard file handles don't have to be opened, and there-
; fore the init code has go into the read/write routines anyway.
; Input:  DS:DX  -  pointer to file name
; Output: AX     -  file handle or error code if carry flag set
;         Carry flag  -  set if error
; Registers changed: AX
;
dos3D		proc	near

	cld
	push	es
	push	di
	push	bx
	mov	es,old_ds		; scan through file name table
	mov	di,dx
	mov	bx,offset cgroup:filetab
dos3D1:	mov	al,cs:[bx]
	cmp	al,0FFh			; check if at end of table
	je	dos3D3
	call	cmpname
	cmp	al,0FFh			; check if filename found
	je	dos3D4
	inc	bx
dos3D2:	inc	bx			; advance BX to the next entry
	mov	al,cs:[bx]		; in the file name table by searching
	or	al,al			; for the end of the current string
	jnz	dos3D2			; and then incrementing BX by 1.
	inc	bx
	jmp	dos3D1

dos3D3:	stc
	mov	ax,ERR_NOFILE		; return with error if name not found
	jmp	short DOS3D9

dos3D4:	and	al,03h
	mov	ah,cs:[bx]		; get device major number
	shl	ah,1
	shl	ah,1			; compute file handle from minor and
	or	al,ah			; major number
	xor	ah,ah
	add	ax,RES_HANDLES
	clc

dos3D9:	pop	bx
	pop	di
	pop	es
	ret

dos3D		endp


;
;**************************************************************************
;
; Close file. There is actually nothing to be done here, since all
; BIOS devices don't have to be cosed.
; Input:  BX  -  file handle
; Output: AX  -  error code if carry flag set
;         Carry flag  -  set if error
; Registers changed: AX
;
dos3E		proc	near

	cmp	bx,MAX_HANDLE
	jbe	dos3E1
	mov	ax,ERR_INVHDL
	stc
dos3E1:	ret

dos3E		endp


;
;**************************************************************************
;
; Read file.
; Input:  BX     -  file handle
;         CX     -  number of bytes to read
;         DS:DX  -  pointer to read buffer
; Output: AX     -  number of bytes read, or error code if carry flag set
;         Carry flag  -  set if error
; Registers changed: AX
;
dos3F		proc	near

	push	bx
	push	di
	push	es

; First check if the file handle is correct, and reset the device if
; necessary.

	cmp	bx,MAX_HANDLE
	jbe	dos3F1
	mov	ax,ERR_INVHDL		; check that the file handle is
	jmp	short dos3F8		; correct
dos3F1:	push	dx
	mov	bh,bl
	and	bl,03h			; compute minor device number
	shr	bh,1
	shr	bh,1			; compute major device number
	cmp	bh,DEV_PRN
	jne	dos3F2			; cannot read from printer
	pop	dx
	mov	ax,ERR_NODEV
	jmp	short dos3F8
dos3F2:	call	devinit			; initialize device if necessary
	pop	dx

; Next read the data from the device into the buffer.

	mov	di,dx
	shr	di,1
	shr	di,1
	shr	di,1			; load buffer address
	shr	di,1
	mov	ax,old_ds
	add	ax,di
	mov	es,ax
	mov	di,dx
	and	di,000Fh
dos3F3:	call	readchar		; read character from device
	jnc	dos3F4
	mov	ax,ERR_READ		; return if error
	jmp	short dos3F8

; Handle end-of-line or end-of-file.

dos3F4:	cmp	al,CHR_CR		; end of line?
	je	dos3F5
	cmp	al,CHR_LF
	je	dos3F5
	cmp	al,CHR_EOF		; or end of file?
	jne	dos3F6
dos3F5:	mov	al,CHR_CR
	call	writechar		; return to a new line on the screen
	mov	al,CHR_LF
	call	writechar
	mov	ax,di
	sub	ax,dx			; compute actual length of input buffer
	clc
	jmp	short dos3F9

; Handle backspace character

dos3F6:	cmp	al,CHR_BS
	jne	dos3F7
	cmp	dx,di			; already at beginning of line
	je	dos3F3
	dec	di
	call	writechar		; print backspace onto the screen
	jmp	dos3F3

; Add character to the end of the line

dos3F7:	cmp	al,' '			; only care for printable characters
	jb	dos3F3
	push	ax
	mov	ax,cx
	add	ax,dx			; check if end of buffer reached
	cmp	ax,di
	pop	ax
	jae	dos3F5
	stosb				; save character into buffer
	call	writechar		; print character
	jmp	dos3F3

dos3F8:	stc
dos3F9:	pop	es
	pop	di
	pop	bx
	ret

dos3F		endp


;
;**************************************************************************
;
; Write to file.
; Input:  BX     -  file handle
;         CX     -  buffer size
;         DS:DX  -  output buffer
; Output: AX     -  number of characters written, or error code if carry set
;         Carry flag  -  set if error
; Registers changed: AX
;
dos40		proc	near

	push	bx
	push	si
	push	ds

; First check if the file handle is correct, and reset the device if
; necessary.

	cmp	bx,MAX_HANDLE
	jbe	dos401
	mov	ax,ERR_INVHDL		; check that the file handle is
	jmp	short dos408		; correct
dos401:	push	dx
	mov	bh,bl
	and	bl,03h			; compute minor device number
	shr	bh,1
	shr	bh,1			; compute major device number
	call	devinit			; initialize device if necessary
	pop	dx

; Next write all characters onto the output device

	mov	si,dx
	shr	si,1
	shr	si,1
	shr	si,1
	shr	si,1			; load buffer address
	mov	ax,old_ds
	add	ax,si
	mov	ds,ax
	mov	si,dx
	and	si,000Fh
dos402:	mov	ax,dx
	add	ax,cx			; check if at end of buffer
	cmp	si,ax
	jae	dos403
	lodsb
	call	writechar		; print character
	jnc	dos402
	mov	ax,ERR_WRITE
	jmp	short dos408
dos403:	mov	ax,si
	sub	ax,dx			; compute number of bytes written
	clc
	jmp	short dos409

dos408:	stc				; return with error
dos409:	pop	ds
	pop	si
	pop	dx
	ret

dos40		endp


;
;**************************************************************************
;
; Read character from device.
; Input:  BH  -  major device number
;         BL  -  minor device number
; Output: AL  -  character
;         Carry flag  -  set if error
; Registers changed: AX
;
readchar	proc	near

	cmp	bh,DEV_CON		; check for keyboard input
	jne	read2
read1:	xor	ah,ah
	int	16h			; read character from keyboard
	or	al,al			; don't accept extended key codes
	jz	read1
	clc
	jmp	short read9

read2:	cmp	bh,DEV_AUX		; check for serial device
	jne	read8
	push	dx
	xor	dh,dh
	mov	dl,bl
	mov	ah,02h			; read character from serial device
	int	14h
	pop	dx
	test	ah,10011110b		; return with error
	jnz	read8
	clc
	jmp	short	read9

read8:	xor	al,al			; illegal major number for reading or
	stc				; other error
read9:	ret

readchar	endp


;
;**************************************************************************
;
; Write character to device.
; Input:  BH  -  major device number
;         BL  -  minor device number
;         AL  -  character to write
; Output: Carry flag  -  set if error
; Registers changed: AX
;
writechar	proc	near

	cmp	bh,DEV_CON		; check for console output
	jne	write1
	push	bx
	mov	ah,0Eh
	mov	bx,0007h		; set color and screen page
	int	10h			; print character
	pop	bx
	clc
	jmp	short write9

write1:	cmp	bh,DEV_AUX		; check for serial device
	jne	write2
	push	dx
	xor	dh,dh
	mov	dl,bl
	mov	ah,01h
	int	14h			; send character to serial port
	pop	dx
	test	ah,10001100b		; check for transmission error
	jnz	write8
	clc
	jmp	short write9

write2:	cmp	ah,DEV_PRN		; check for printer device
	jne	write8
	push	dx
	xor	dh,dh
	mov	dl,bl
	xor	ah,ah
	int	17h			; send character to printer
	pop	dx
	test	ah,00101001b		; check for error
	jnz	write8
	clc
	jmp	short write9

write8:	stc				; return with error
write9:	ret

writechar	endp


;
;**************************************************************************
;
; Compare two device strings.
; Input:  CS:BX  -  pointer to device table entry
;         ES:DI  -  pointer to zero terminated file name
; Output: AL     -  minor number, or 0FFh if strings not equal
; Registers changed: AX
;
cmpname		proc	near

	push	bp
	push	si
	mov	bp,di
	xor	si,si			; index pointer into string

cmpn1:	mov	al,cs:[bx + si + 2]	; get next character from both strings
	mov	ah,es:[bp + si]
	or	al,al
	jnz	cmpn2
	or	ah,ah			; if at end of both strings, they are
	jz	cmpn7			; equal
	jmp	short cmpn8		; if just one string is at the end,
cmpn2:	or	ah,ah			; both strings are not equal
	jz	cmpn8

	cmp	al,'?'			; a wildcard character means catch all
	je	cmpn3
	cmp	al,ah			; if characters are not equal, both
	jne	cmpn8			; strings are not equal
cmpn3:	inc	si			; advance to next character
	jmp	cmpn1

cmpn7:	xor	ah,ah
	mov	al,cs:[bx + 1]		; get position of minor number in name
	cmp	al,0FFh			; if no minor number, assume 0
	je	cmpn9
	mov	si,2
	add	si,ax			; get minor number from string
	mov	ah,es:[bp + si]
	sub	ah,'0'			; convert it into a binary number
	jb	cmpn8			; if invalid return not found
	cmp	ah,MAX_MINOR
	jbe	cmpn9			; got a valid minor number
cmpn8:	mov	ah,0FFh
cmpn9:	mov	al,ah
	pop	si
	pop	bp
	ret

cmpname		endp


;
;**************************************************************************
;
text_end

	end

