;**************************************************************************
;*
;* Boot-ROM-Code to load an operating system across a TCP/IP network.
;*
;* Module:  process.asm
;* Purpose: Process management functions for DOS simulator
;* Entries: dos00, dos26, dos31, dos4B, dos4C, dos4D
;*
;**************************************************************************
;*
;* Copyright (C) 1995 Gero Kuhlmann <gero@gkminix.han.de>
;*
;*  This program is free software; you can redistribute it and/or modify
;*  it under the terms of the GNU General Public License as published by
;*  the Free Software Foundation; either version 2 of the License, or
;*  any later version.
;*
;*  This program is distributed in the hope that it will be useful,
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;*  GNU General Public License for more details.
;*
;*  You should have received a copy of the GNU General Public License
;*  along with this program; if not, write to the Free Software
;*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;*


;
;**************************************************************************
;
; Include assembler macros:
;
include ..\headers\asm\layout.inc
include ..\headers\asm\macros.inc
include ..\headers\asm\memory.inc
include ..\headers\asm\minidos.inc
include .\dospriv.inc


;
;**************************************************************************
;
; Data segment
;
data_start

		public	curr_psp

		extrn	dos_active:byte	; DOS active flag
		extrn	old_ds:word	; caller's DS register
		extrn	old_stack:dword	; stack pointer of calling process

ret_code	db	0		; program return code
ret_cause	db	0		; program return cause
curr_psp	dw	0		; current PSP
psp_tmpl	psp	<>		; PSP template

data_end


;
;**************************************************************************
;
; Start code segment.
;
text_start

	public	dos00, dos26, dos31		; define entry poimts
	public	dos4B, dos4C, dos4D

	extrn	call21:far
	extrn	dos48:near
	extrn	dos49:near
	extrn	dos4A:near
	extrn	freeall:near
	extrn	dosfatal:near


; Magic cookie to identify COM program images

cookie	db	COM_MAGIC



;
;**************************************************************************
;
; Terminate program.
; Input:  none
; Output: none (function does not return)
; Registers changed: all
;
dos00		proc	near

	les	bx,old_stack
	mov	dx,es:[bx + 2]		; load caller's code segment
	mov	curr_psp,dx
	xor	al,al
	jmp	dos4C

dos00		endp


;
;**************************************************************************
;
; Setup new PSP.
; Input:  DX  -  Segment of new PSP
; Output: none
; Registers changed: none
;
dos26		proc	near

	push	ax
	push	es
	mov	es,dx
	call	newpsp
	pop	es
	pop	ax
	ret

dos26		endp


;
;**************************************************************************
;
; Terminate current process, but keep it resident in memory.
; Input:  AL  -  return code
;         DX  -  number of paragraphs to keep in memory
; Output: none (function does not return)
; Registers changed: all
;
dos31		proc	near

	push	ax
	mov	ax,curr_psp
	or	ax,ax			; is there a program running at all?
	jnz	dos311
	mov	ax,FATAL_NOPROC		; nope, can't continue
	jmp	dosfatal

dos311:	mov	bx,dx
	mov	dx,ax
	dec	ax			; compute MCB address
	mov	es,ax
	mov	al,es:[mcb_id]		; must be valid ID
	cmp	al,MCB_MEM_ID
	je	dos312
	cmp	al,MCB_END_ID
	jne	dos313

dos312:	mov	word ptr es:[mcb_owner],0FFFFh	; this saves the block from
	call	freeall				; being freed
	jnc	dos314
dos313:	mov	ax,FATAL_MEMORY			; unable to continue
	jmp	dosfatal

dos314:	mov	word ptr es:[mcb_owner],dx	; restore PSP address in MCB
	mov	es,dx
	call	dos4A				; resize the memory block
	jc	dos313

	pop	ax
	mov	ret_code,al			; set return code and terminate
	mov	ret_cause,3
	jmp	terminate

dos31		endp


;
;**************************************************************************
;
; Start a new process. In the real DOS this function reads a program
; from disk and then starts it. Since this DOS simulator does not
; support disk files, we have to read the program from somewhere
; else, e.g. memory. Instead of a program name, the parent process
; provides the memory address of the file image. Note that this
; image must reside outside of the memory space which is used and
; maintained by the DOS simulator. Also note, that the simulator can
; only handle COM-type programs, not EXE-type programs.
; The program image must begin with a magic cookie followed by it's
; size in paragraphs.
; Input:  AL     -  must be 0
;         ES:BX  -  pointer to parameter block
;         DS:DX  -  pointer to program code image
; Output: AX  -  error code if carry flag set
;         Carry flag  -  set if error
; Registers changed: all
;
dos4B		proc	near

	cld
	push	es
	push	bx
	or	al,al
	mov	ax,ERR_NOFN		; overlay loading is not supported
	jnz	dos4B9

; First check if the program code image has the correct magic cookie
; at it's beginning.

	push	ds
	mov	es,old_ds
	mov	di,dx
	mov	ax,cs
	mov	ds,ax
	mov	si,offset cgroup:cookie
	mov	cx,COM_MAGIC_LEN
	repe	cmpsb			; check for magic cookie
	pop	ds
	jz	dos4B1
	mov	ax,ERR_INVFORM		; error: invalid format
	jmp	short dos4B9

; Next allocate 64 kB of memory for the program, or as much as possible
; if 64 kB are not available. Then check if the program will fit into
; that memory space together with the PSP and some safety margin for
; the stack.

dos4B1:	mov	bx,1000h		; try to acquire 64 kB of memory
	call	dos48
	jnc	dos4B2
	cmp	ax,ERR_NOMEM
	jne	dos4B9			; couldn't get that much, so try
	call	dos48			; to get as much as possible
	jc	dos4B9
dos4B2:	sub	bx,COM_MIN_FREE
	jbe	dos4B3			; check if the program will fit into
	cmp	es:[di],bx		; the reserved memory area
	jb	dos4B4
dos4B3:	mov	es,ax
	call	dos49			; it doesn't fit, so free all memory
	jc	dos4B9
	mov	ax,ERR_NOMEM		; and return with error
dos4B9:	pop	bx
	pop	es
	stc
	ret

; Now setup the PSP and copy the program code into the new memory block. Also
; compute the offset of the first usable memory address after the end of the
; program, and save it into the PSP.

dos4B4:	mov	es,ax
	call	newpsp			; create new PSP
	push	ds
	mov	ds,old_ds
	mov	si,dx
	mov	ax,[COM_SIZE_OFS + si]
	mov	cl,3			; size is in paragraphs, convert it
	shl	ax,cl			; to number of words
	mov	cx,ax
	add	si,COM_CODE_OFS		; copy program code image into new
	mov	di,PSP_SIZE		; place
	rep	movsw
	pop	ds

	mov	ax,di
	add	ax,0Fh
	mov	cl,4			; compute segment of first usable
	shr	ax,cl			; address after the program end
	mov	cx,es
	add	ax,cx
	mov	es:[psp_endmem],ax

; Next copy the command line into the PSP. The FCB's remain empty because we
; don't use FCB's anyway.

	pop	bx
	pop	dx
	push	ds
	mov	ds,dx
	mov	ax,ds:prog_env[bx]	; save pointer to environment
	mov	es:[psp_env],ax
	lds	si,prog_cmdl[bx]	; get pointer to command line
	mov	ax,ds
	or	ax,si
	jz	dos4B5
	mov	cl,ds:[si]		; save size of command line
	mov	es:[psp_cmdlsize],cl
	xor	ch,ch
	jcxz	dos4B5
	cmp	cx,PSP_MAXCMDL
	jbe	dos4B8
	mov	cx,PSP_MAXCMDL
dos4B8:	inc	si
	mov	di,psp_cmdl
	rep	movsb			; save command line
	mov	al,0Dh			; terminate command line with CR
	stosb
	xor	al,al			; and with a zero byte
	stosb

; The PSP has now been setup. Now set the owner of the program's memory
; block and switch the stack to the new program.

dos4B5:	mov	ax,es
	dec	ax
	mov	ds,ax			; get size of memory block from
	mov	ax,ds:[mcb_size]	; MCB
	mov	ds:[mcb_owner],es	; set owner of memory block
	pop	ds
	mov	cl,4
	shl	ax,cl			; convert paragraph size into byte
	dec	ax			; offset for top of stack
	mov	dx,es
	cli
	mov	ss,dx			; set new stack
	mov	sp,ax

; Finally reset the DOS active byte, make the new program current, and call
; it with the return pointer set correctly on the new stack.

	mov	dos_active,0		; reset DOS active byte
	mov	curr_psp,dx		; save new PSP pointer
	mov	ds,dx			; set data segment register correctly
	xor	ax,ax
	push	ax			; save near pointer to terminate fctn
	push	es
	mov	ax,PSP_SIZE		; save pointer to the new program's
	push	ax			; entry point
	sti
	retf				; call the new program

dos4B		endp


;
;**************************************************************************
;
; Terminate current program.
; Input:  AL  -  return code
; Output: none (function does not return)
; Registers changed: all
;
dos4C		proc	near

	push	ax
	mov	ax,curr_psp
	or	ax,ax			; is there a program running at all?
	jnz	dos4C1
	mov	ax,FATAL_NOPROC		; nope, can't continue
	jmp	dosfatal

dos4C1:	mov	dx,ax
	dec	ax			; compute MCB address
	mov	es,ax
	mov	al,es:[mcb_id]		; must be valid ID
	cmp	al,MCB_MEM_ID
	je	dos4C2
	cmp	al,MCB_END_ID
	jne	dos4C3

dos4C2:	call	freeall			; free all memory used by this process
	jc	dos4C3
	or	cx,cx			; at least one block (PSP) has to be
	jnz	dos4C4			; freed
dos4C3:	mov	ax,FATAL_MEMORY		; unable to continue
	jmp	dosfatal

dos4C4:	pop	ax
	mov	ret_code,al
	mov	ret_cause,0
	jmp	short terminate

dos4C		endp


;
;**************************************************************************
;
; Get exit code.
; Input:  none
; Output: AL  -  exit code
;         AH  -  exit cause
; Registers changed: AX
;
dos4D		proc	near

	mov	al,ret_code
	mov	ah,ret_cause
	ret

dos4D		endp


;
;**************************************************************************
;
; Terminate current process. This routine does the actual job, except
; memory management.
; Input:  none
; Output: none (routine does not return)
; Registers changed: all
;
terminate	proc	near

	cld

; First restore all the interrupt vectors from the PSP, and then call
; the termination handler.

	push	ds
	mov	dx,curr_psp
	mov	ds,dx
	xor	ax,ax
	mov	es,ax
	mov	si,psp_int22
	mov	di,22h * 4
	mov	cx,6
	rep	movsw			; restore INT's 22h, 23h and 24h
	mov	es,dx
	int	22h			; call the termination handler
	pop	ds

; Next restore the stack of the parent process, reset the PSP of the
; parent process, reset the DOS active flag, and return to the parent
; calling address.

	cli
	mov	es,curr_psp
	mov	dos_active,0		; reset DOS active flag
	mov	ax,es:[psp_parent]	; reset current PSP
	mov	curr_psp,ax
	mov	ss,word ptr es:[psp_oldstack+2]
	mov	sp,word ptr es:[psp_oldstack+0]
	xor	ax,ax			; no return error
	clc				; must return with carry cleared
	iret				; return to caller, also restores
					; the interrupt flag
terminate	endp


;
;**************************************************************************
;
; Create new PSP.
; Input:  ES  -  Segment of new PSP
; Output: none
; Registers changed: AX
;
newpsp		proc	near

	cld
	push	cx
	push	si
	push	di

; first clear the whole PSP area, and then copy the PSP template
; into the new PSP area.

	xor	ax,ax
	xor	di,di
	mov	cx,PSP_SIZE
	rep	stosb

	xor	di,di
	mov	si,offset dgroup:psp_tmpl
	mov	cx,size psp_tmpl
	rep	movsb

; next copy all interrupt vectors and call addresses

	mov	ax,ds			; prevent pushing another value
	xor	cx,cx			; onto stack, e.g. take care for
	mov	ds,cx			; stack overflows.
	mov	si,22h * 4
	mov	di,psp_int22		; copy interrupt vectors
	mov	cx,6
	rep	movsw
	mov	word ptr es:[psp_call21 + 0],offset cgroup:call21
	mov	word ptr es:[psp_call21 + 2],cs
	mov	ds,ax

; copy some internal data into the PSP

	mov	es:psp_seghndl,es
	mov	ax,curr_psp
	mov	es:psp_parent,ax
	mov	ax,word ptr [old_stack + 0]
	mov	word ptr es:[psp_oldstack + 0],ax
	mov	ax,word ptr [old_stack + 2]
	mov	word ptr es:[psp_oldstack + 2],ax

	pop	di
	pop	si
	pop	cx
	ret

newpsp		endp


;
;**************************************************************************
;
text_end

	end

