;**************************************************************************
;*
;* Boot-ROM-Code to load an operating system across a TCP/IP network.
;*
;* Module:  floboot.asm
;* Purpose: Floppy boot sector to load the rest of the boot rom code
;* Entries: called by BIOS at offset 0
;*
;**************************************************************************
;*
;* Copyright (C) 1995 Gero Kuhlmann <gero@gkminix.han.de>
;*
;*  This program is free software; you can redistribute it and/or modify
;*  it under the terms of the GNU General Public License as published by
;*  the Free Software Foundation; either version 2 of the License, or
;*  any later version.
;*
;*  This program is distributed in the hope that it will be useful,
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;*  GNU General Public License for more details.
;*
;*  You should have received a copy of the GNU General Public License
;*  along with this program; if not, write to the Free Software
;*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;*


;
;**************************************************************************
;
; Load the boot rom segment layout and assembler macros:
;
include ..\headers\asm\layout.inc
include ..\headers\asm\macros.inc
include ..\headers\asm\memory.inc


;
;**************************************************************************
;
; Equates local for this module:
;
STACKTOP	equ	0FFFEh		; top of stack in code segment
HIGHADR		equ	MEMEND		; end of physical base memory
LOADADR		equ	HIGHMEM		; segment where to load the boot code
LOADMEM		equ	(HIGHADR-LOADADR)*16	; required load memory size
DPTVECTOR	equ	0078h		; address of pointer to DPT
SECTSIZE	equ	512		; size of one floppy sector


;
;**************************************************************************
;
; Structure defining the layout of a disk parameter table
;
dptab		struc

dp_spec1	db	0
dp_spec2	db	0
dp_moffc	db	0
dp_byt_sec	db	0	; bytes per sector
dp_sec_track	db	0	; sectors per track
dp_gap		db	0
dp_dtl		db	0
dp_gap3		db	0
dp_fil_byte	db	0
dp_hd_time	db	0
dp_most		db	0
dp_maxtrack	db	0
dp_other	db	0

dptab		ends


;
;**************************************************************************
;
; Here the loader program starts. Some ideas are taken from the Linux
; boot sector. Note that this program has to be completely seperate
; from all other segments of the boot rom code. It can contain data
; since it's getting loaded into the ram.
; The loader works by first patching the floppy disk parameter table,
; and then trying to read the last sector of the floppy in order to
; find out about the disk size. Then it will load 32 kB from the
; diskette just below the end of physical ram. A problem exists
; with those systems, whose BIOS reserves itself 1 or 2 kB at that
; memory area. I don't know what to do in this case. At present the
; boot loader issues a warning message and then just stops reading
; when the beginning of the extended BIOS area is reached. Hopefully
; there is no further code or data which is missing at execution
; time.

load_start				; begin loader text segment

	extrn	start:far		; this will include the rom code

	org	0

	cli
	call	near ptr dobot1
dobot1:	pop	ax			; determine the offset
	sub	ax,offset load_text:dobot1

	mov	cl,4
	mov	bx,cs
	mov	dx,bx			; convert the base address from
	and	dx,0F000h		; seg:ofs format into sort of a
	shl	bx,cl			; linear address. this conversion
	add	bx,ax			; assumes that no overflow can occur.

	mov	ax,bx			; compute a code segment address
	shr	ax,cl			; which lets this code start at
	add	ax,dx			; offset 0. this is necessary so
	push	ax			; all offsets computed by masm work.
	mov	ax,offset load_text:dobot2
	push	ax
	retf				; this will jump to dobot2 with new
					; code segment register
dobot2:	mov	ax,cs
	mov	ds,ax			; load segment registers
	mov	es,ax
	mov	bx,STACKTOP
	mov	ss,ax
	mov	sp,bx			; finally set the stack
	sti

; now it's time to print some really important messages :-))

	mov	si,offset load_text:sigmsg
	call	prnstr

; After all segment registers have been set, find out about the RAM
; size next and determine how much memory is available in the loading
; area. If it's not enough, issue a warning message, wait for a key
; press and then continue. It's the users responsibility to act
; correctly :-))

	int	12h			; get memory size in kB
	mov	si,offset load_text:memerr
	mov	cl,6
	ror	ax,cl			; convert it into bytes
	mov	dx,ax
	and	dx,003FFh		; get most significant word
	and	ax,0FC00h		; get least significant word
	sub	ax,(LOADADR shl 4)	; compute the number of bytes available
	sbb	dx,(LOADADR shr 12)	; in the loading area
	jnc	memok1
	jmp	reboot

memok1:	or	dx,dx
	mov	dx,LOADMEM
	jnz	memok2
	cmp	ax,LOADMEM		; we have enough memory, so just load
	jae	memok2			; LOADMEM number of bytes
	mov	dx,ax			; otherwise load as much as possible
	call	prnstr			; print memory warning
	mov	si,offset load_text:keyprs
	call	waitk
memok2:	mov	ldsize,dx		; save amount of load memory

; We now know that loading is possible. Next determine the number of
; sectors per track on the floppy. To do this the disk parameter
; table has first to be changed, so that the BIOS is able to access
; at least 18 sectors in a track.

	cld
	push	ds
	mov	dx,ds
	xor	ax,ax
	mov	es,ax			; first save the old disk parameter
	lds	si,es:[DPTVECTOR]	; table
	mov	es,dx
	mov	di,offset load_text:newdpt
	mov	cx,size dptab
	rep	movsb
	pop	ds
	mov	al,18
	mov	newdpt.dp_sec_track,al	; set new number of sectors per track

	xor	ax,ax			; set pointer to new DPT
	mov	es,ax
	mov	es:[DPTVECTOR + 0],offset load_text:newdpt
	mov	es:[DPTVECTOR + 2],ds

; Next try to find the number of sectors per track. To do this, the
; last sector for each disk format is read. This idea has been taken
; from the Linux boot sector.

	xor	ah,ah			; reset drive 0
	xor	dl,dl
	int 	13h

	mov	ax,LOADADR
	mov	es,ax			; set load address 9800:0
	xor	bx,bx

	xor	dx, dx			; drive 0, head 0
	mov	cx,18			; sector 18, track 0
	mov	ax,0201h		; read one sector
	int	13h
	jnc	gotit
	mov	cx,15			; sector 15
	mov	ax,0201h		; read one sector
	int	13h
	jnc	gotit
	mov	cx,9
gotit:	mov	sects,cx		; save number of sector per track

; Now load the next 32 kB behind this first sector. Note that if we
; don't have enough memory available, just load as many sectors as
; will fit.

	mov	ax,LOADADR
	mov	es,ax			; setup load segment
	xor	bx,bx			; starting offset of load memory
	mov	dx,ldsize
	mov	cl,9			; convert number of bytes into
	shr	dx,cl			; number of sectors
	mov	ax,1			; starting sector
rdloop:	or	dx,dx
	jz	loaded
	push	ax
	push	dx
	call	cvtsec			; compute sector information
	mov	ax,0201h		; load one sector
	int	13h
	pop	dx
	pop	ax
	jc	lderr
	dec	dx			; decrement sector count
	inc	ax			; increment sector number
	add	bx,SECTSIZE		; increment load pointer
	jmp	rdloop

lderr:	mov	si,offset load_text:dskerr
	jmp	short reboot

; The boot code is in the memory now. It has to be called exactly
; like the BIOS would call it in a real rom. First just call the
; entry routine, which should setup int 19h. Then call the boot-
; strap vector. It should never return.

loaded:	push	cs
	mov	ax,offset load_text:boot
	push	ax
	mov	ax,LOADADR
	push	ax
	mov	ax,offset cgroup:start
	push	ax
	retf				; call boot rom startup code

reboot:	call	prnstr
	mov	si,offset load_text:botmsg
	call	waitk
boot:	int	19h			; reboot the system
@@:	sti
	hlt				; should never happen
	jmp	@B


;
;**************************************************************************
;
; Routine to convert an absolute sector into a sector/track/head information.
; Input:  AX  -  absolute sector number
; Output: CL  -  sector number
;         CH  -  track number
;         DH  -  head number
;         DL  -  0 (number for drive 0)
; Registers changed: AX, CX, DX
;
cvtsec	proc	near

	xor	dx,dx
	div	sects			; compute sector number in a track
	inc	dl			; sector numbers always start with 1
	mov	cl,dl
	mov	dx,ax
	shr	ax,1			; divide by number of heads
	mov	ch,al			; save track number
	and	dx,1			; mask head number
	xchg	dl,dh
	ret

cvtsec	endp


;
;**************************************************************************
;
; Routine to print a string onto the console.
; Input:  DS:SI  -  address of NULL terminated string
; Output: None
; Registers changed: AX, SI
;
prnstr	proc	near

prnst1:	lodsb				; loop over all characters
	or	al,al
	jz	prnst2
	call	prnchr			; send character to screen
	jmp	prnst1
prnst2:	ret

prnstr	endp


;
;**************************************************************************
;
; Routine to print a character onto the console.
; Input:  AL  -  character
; Output: None
; Registers changed: AX
;
prnchr	proc	near

	push	bx
	mov	ah,0Eh			; BIOS command for character output
	mov	bx,0007h		; screen page and colors
	int	10h			; send character to screen
	pop	bx
	ret

prnchr	endp


;
;**************************************************************************
;
; Routine to wait for a key press
; Input:  DS:SI  -  Address of string to print before waiting for key press
; Output: None
; Registers changed: AX, SI
;
waitk	proc	near

	call	prnstr
	xor	ah,ah
	int	16h
	mov	si,offset load_text:crlf
	call	prnstr
	ret

waitk	endp


;
;**************************************************************************
;
; Strings and messages:
;
sigmsg	db	'Disk loader for net boot'
	db	0Dh,0Ah,0Dh,0Ah,0

dskerr	db	'Error reading diskette'
	db	0Dh,0Ah,0

memerr	db	'Not enough base memory'
	db	0Dh,0Ah,0

botmsg	db	'Rebooting. '
keyprs	db	'Press a key to continue',0

crlf	db	0Dh,0Ah,0


;
;**************************************************************************
;
; Data space:
;
ldsize		dw	0	; amount of memory available for loading
sects		dw	0	; number of sectors per track
newdpt		dptab	<>	; changed disk parameter table

		db	(offset load_text:[SECTSIZE-2] - $) dup (0)
		dw	0AA55h


;
;**************************************************************************
;
load_end			; end load segment

		end

