;**************************************************************************
;*
;* Boot-ROM-Code to load an operating system across a TCP/IP network.
;*
;* Module:  rom.asm
;* Purpose: Setup the stack and data areas, and then call the higher
;*          level routines.
;* Entries: Called by BIOS at offset 0.
;*
;**************************************************************************
;*
;* Copyright (C) 1995 Gero Kuhlmann <gero@gkminix.han.de>
;*
;*  This program is free software; you can redistribute it and/or modify
;*  it under the terms of the GNU General Public License as published by
;*  the Free Software Foundation; either version 2 of the License, or
;*  any later version.
;*
;*  This program is distributed in the hope that it will be useful,
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;*  GNU General Public License for more details.
;*
;*  You should have received a copy of the GNU General Public License
;*  along with this program; if not, write to the Free Software
;*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;*


;
;**************************************************************************
;
; Load the boot rom segment layout and assembler macros.
;
INITSEG		equ	1		; create end-of-segment labels

include ..\headers\asm\layout.inc
include ..\headers\asm\macros.inc
include ..\headers\asm\memory.inc


;
;**************************************************************************
;
; Equates for this module:
;
DATAOFS		equ	8		; Offset of special data values
					; from beginning of rom
BOOTSTRAP	equ	19h		; Bootstrap interrupt vector
OLDBOOT		equ	78h		; Vector for old bootstrap int

STACKSIZE	equ	4096		; Size of stack


;
;**************************************************************************
;
; The physical rom starts here. At the beginning there is some
; miscellaneous data.
;
text_start

	public	start

	extrn	prnstr:near

	org	0			; the rom always starts at offset 0

	dw	0AA55h			; rom signature
	db	ROMSIZE / 512		; size
start:	jmp	short romstart		; the BIOS jumps to this address

	db	(_text:DATAOFS - $) dup (0)	; Fill up

; The following values get filled in by the binary patch program which
; concatenates the boot rom with the packet driver.

pktadr	dw	0			; Address of packet driver 
pktarg	dw	0			; Address of packet driver command line

; This value can be used by the binary patch program to strip off the
; BSS segment from the final binary. The BSS segment just contains zeroes
; which don't have to be saved in the tight ROM.

	dw	offset cgroup:_end_of_data


;
;**************************************************************************
;
; Now start with the actual boot rom code. The startup code just installs
; a new bootstrap interrupt vector, and then returns to the BIOS. This is
; necessary so that the BIOS gets a chance to also initialize other installed
; roms. However, only do this when there is no diskette in drive A. Other-
; wise the new vector does not get installed, which means that the system
; will be booted from diskette.
;
romstart	proc	near

	sti
	mov	ax,cs
	cmp	ax,MEMEND	; only check for disk if we are not booting
	jb	nodisk		; from it :-))
	push	es
	mov	bx,NEWDATA	; should be empty space
	xor	ax,ax		; check if a disk is inserted in drive A:
	xor	dx,dx		; use head 0, drive 0
	int	13h		; initialize drives
	mov	es,bx
	xor	bx,bx
	mov	cx,0001h	; read sector 1 of track 0
	mov	ax,0201h	; read 1 sector
	int	13h
	pop	es
	jc	nodisk
	retf			; return to BIOS to boot from diskette

nodisk:	cli			; do not allow interrupts when changing the
	push	ds		; interrupt vector table
	xor	ax,ax		; copy old interrupt vector
	mov	ds,ax
	mov	ax,ds:[BOOTSTRAP*4 + 0]
	mov	ds:[OLDBOOT*4 + 0],ax
	mov	ax,ds:[BOOTSTRAP*4 + 2]
	mov	ds:[OLDBOOT*4 + 2],ax
	mov	word ptr ds:[BOOTSTRAP*4 + 0],offset cgroup:int19
	mov	word ptr ds:[BOOTSTRAP*4 + 2],cs
	pop	ds
	sti			; that's it
	retf

romstart	endp


;
;**************************************************************************
;
; The bootstrap interrupt routine starts the higher level code after
; copying the data segment into the lower ram area. It will never
; return to the caller.
; The stack segment will be placed right above the BSS segment.
;
int19		proc	far

	assume	ds:cgroup
	assume	es:dgroup

	extrn	_main:near

	cli
	mov	bx,NEWDATA
	mov	dx,offset dgroup:_end_of_bss + STACKSIZE
	mov	ss,bx
	mov	sp,dx		; set new stack pointer

	int	12h		; check if we have enough base memory
	cmp	ax,HIGHMEM / 64	; this error will never happen with the disk
	jae	memok		; boot because it checked for enough memory
	sti			; already
	mov	bx,offset cgroup:memerr
	call	prnstr
	int	OLDBOOT
dolop:	hlt			; should never happen
	jmp	short dolop

memok:	cld
	xor	ax,ax
	mov	ds,ax
	mov	ax,OLDINTS
	mov	es,ax		; save old interrupt vectors and also
	xor	si,si		; restore the bootstrap vector within
	xor	di,di		; the copy of the vector table
	mov	cx,VECTSIZE / 2
	rep	movsw
	mov	ax,word ptr ds:[OLDBOOT*4 + 0]
	mov	word ptr es:[BOOTSTRAP*4 + 0],ax
	mov	ax,word ptr ds:[OLDBOOT*4 + 2]
	mov	word ptr es:[BOOTSTRAP*4 + 2],ax

	mov	ax,cs
	mov	ds,ax		; move data segment into lower ram
	mov	es,bx
	mov	si,offset cgroup:_end_of_code + 1
	and	si,0FFFEh
	mov	di,si		; start the data segment at the right offset
	and	di,000Fh
	mov	cx,offset dgroup:_end_of_data
	mov	ax,cx
	rep	movsb

	mov	cx,offset dgroup:_end_of_bss
	sub	cx,ax
	xor	al,al
	rep	stosb		; clear BSS segment

	mov	es,bx		; finally set the new data segment
	mov	ds,bx	
	sti

; The main routine of the boot rom gets three farpointers onto the
; stack: a pointer to the end of the data area and the two pointers
; at the beginning of the rom code.

	push	ds
	add	dx,2
	push	dx		; push the initial stack pointer
	push	cs
	push	cs:pktarg	; push the packet driver command line pointer
	push	cs
	push	cs:pktadr	; push the packet driver address
	call	_main
	jmp	dolop		; just halt the system (should never happen)


; Error message if not enough base memory found. Not a very long message to
; save memory space. Should never happen anyway...

memerr	db	0Dh,0Ah
	db	'Bootrom: memerr'
	db	0

int19		endp


;
;**************************************************************************
;
text_end			; end text segment

	end
