;**************************************************************************
;*
;* Boot-ROM-Code to load an operating system across a TCP/IP network.
;*
;* Module:  poll.asm
;* Purpose: Setup a list of routines to poll during waiting states (keyboard
;*          etc.). This module is required for handling of hardware interrupts.
;* Entries: _setpoll, _resetpoll, _markpoll, _dopoll
;*
;**************************************************************************
;*
;* Copyright (C) 1995 Gero Kuhlmann <gero@gkminix.han.de>
;*
;*  This program is free software; you can redistribute it and/or modify
;*  it under the terms of the GNU General Public License as published by
;*  the Free Software Foundation; either version 2 of the License, or
;*  any later version.
;*
;*  This program is distributed in the hope that it will be useful,
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;*  GNU General Public License for more details.
;*
;*  You should have received a copy of the GNU General Public License
;*  along with this program; if not, write to the Free Software
;*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;*


;
;**************************************************************************
;
; Include assembler macros:
;
include ..\headers\asm\layout.inc
include ..\headers\asm\macros.inc
include ..\headers\asm\memory.inc
include .\libpriv.inc


;
;**************************************************************************
;
; Structure for each element of the polling table:
;
poll_item	struc

poll_active	db	-1		; -1 = unused, 0 = inactive, 1 = active
poll_event	dw	0		; address of event handler routine

poll_item	ends

; Define maximum number of entries in the polling table. Since there are
; not many hardware interrupts expected (which could trigger the polling
; mechanism), this table can be fairly small.

EVENTNUM	equ	4		; maximum number of event handlers


;
;**************************************************************************
;
; Data segment:
;
data_start

polltab		poll_item	EVENTNUM dup (<>)
ispolling	db		0	; flag indicating active polling

data_end


;
;**************************************************************************
;
; Start code segment.
;
text_start

	public	_setpoll		; define entry points
	public	_resetpoll
	public	_markpoll
	public	_dopoll


;
;**************************************************************************
;
; Set an event handler in the polling table.
; Input:  1. arg  -  short pointer to event handler routine
; Output: int     -  handle number
;
_setpoll	proc	near

	push	bp
	mov	bp,sp			; setup standard stack frame

	mov	bx,offset dgroup:polltab
	mov	cx,length polltab
setp1:	mov	al,poll_active[bx]
	cmp	al,-1			; find a free table entry
	je	setp2
	add	bx,size poll_item
	loop	setp1
	mov	ax,0FFFFh		; no free entry found, return error
	jmp	short setp9

setp2:	getarg	ax,0			; save pointer to event handler in
	mov	poll_event[bx],ax	; poll table
	mov	poll_active[bx],0	; and mark it used
	mov	ax,length polltab	; compute handle number
	sub	ax,cx

setp9:	pop	bp
	ret

_setpoll	endp


;
;**************************************************************************
;
; Mark an event handler as unused.
; Input:  1. arg  -  handle number into poll table
; Output: none
;
_resetpoll	proc	near

	push	bp
	mov	bp,sp
	getarg	bx,0			; get handle number
	mov	al,-1			; mark entry unused
	call	setactive
	pop	bp
	ret

_resetpoll	endp


;
;**************************************************************************
;
; Mark an event handler as active.
; Input:  1. arg  -  handle number into poll table
; Output: none
;
_markpoll	proc	near

	push	bp
	mov	bp,sp
	getarg	bx,0			; get handle number
	mov	al,1			; mark entry as active
	call	setactive
	pop	bp
	ret

_markpoll	endp


;
;**************************************************************************
;
; Loop through poll table and execute every event handler which is
; marked busy. The handler routines are not allowed to change any
; registers except AX, BX, CX, DX, SI and DI.
; Input:  none
; Output: none
; Registers changed: AX
;
_dopoll		proc	near

	push	bx
	push	cx
	test	ispolling,0FFh		; check if we are already polling
	jnz	dopol9
	inc	ispolling		; set activity flag
	mov	bx,offset dgroup:polltab
	mov	cx,length polltab
dopol1:	mov	al,poll_active[bx]
	dec	al			; is entry active?
	jnz	dopol2
	push	bx
	push	cx
	push	dx
	push	si
	push	di
	call	poll_event[bx]		; call event handler
	pop	di
	pop	si
	pop	dx
	pop	cx
	pop	bx
	or	ax,ax			; don't reset the flag if the
	jnz	dopol2			; handler doesn't give it's OK
	mov	poll_active[bx],al	; reset active byte
dopol2:	add	bx,size poll_item	; loop through whole table
	loop	dopol1
	dec	ispolling		; reset activity flag
dopol9:	pop	cx
	pop	dx
	ret

_dopoll		endp


;
;**************************************************************************
;
; Set active byte in poll table entry.
; Input:  BX  -  number of poll table entry
;         AL  -  new value of active flag
; Output: none
; Registers changed: AX, BX
;
setactive	proc	near

	cmp	bl,length polltab		; valid?
	jae	seta9
	mov	bh,al
	mov	al,size poll_item		; compute offset into poll table
	mul	bl
	xchg	bx,ax
	mov	polltab[bx].poll_active,ah	; set active flag
seta9:	ret

setactive	endp


;
;**************************************************************************
;
text_end

	end

