/*
 **************************************************************************
 *
 * Boot-ROM-Code to load an operating system across a TCP/IP network.
 *
 * Module:  arp.c
 * Purpose: ARP handling
 * Entries: init_arp, resolve, addcache
 *
 **************************************************************************
 *
 * Copyright (C) 1995 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "../headers/packet.h"
#include "../headers/net.h"
#include "../headers/romlib.h"
#include "../headers/general.h"
#include "./netpriv.h"
#include "./arp.h"



/*
 **************************************************************************
 * 
 * Global variables
 */
static unsigned char *arpbuf = NULL;		/* ARP send buffer	*/
static struct arp_cache *arp_list = NULL;	/* ARP address cache	*/
static int cache_count = 1;			/* Cache timeout cnt	*/



/*
 **************************************************************************
 * 
 * Find an IP address in the ARP cache. Return -1 if not found.
 */
static int findip(t_ipaddr ip)
{
  int i;

  for (i = 0; i < ARPNUM; i++)
	if (arp_list[i].usecnt > 0 && arp_list[i].ipaddr == ip)
		return(i);
  return(-1);
}



/*
 **************************************************************************
 * 
 * Add a new entry to the ARP cache.
 */
void addcache(unsigned char far *ha, t_ipaddr ip)
{
#ifndef NOARP
  int i, oldest, oldval;

  /*
   * First search for the oldest entry in the cache. Also delete all
   * old entries with the current IP address.
   */
  oldest = 0;
  oldval = cache_count;
  for (i = 1; i < ARPNUM; i++) {
	if (arp_list[i].ipaddr == ip)
		arp_list[i].usecnt = 0;
	if (arp_list[i].usecnt < oldval) {
		oldest = i;
		oldval = arp_list[i].usecnt;
	}
  }

  /* Then put the new entry into the cache */
  memcpy(arp_list[oldest].hwaddr, ha, ETH_ALEN);
  arp_list[oldest].ipaddr = ip;
  arp_list[oldest].usecnt = cache_count++;
#else
  memcpy(arp_list[0].hwaddr, ha, ETH_ALEN);
  arp_list[0].ipaddr = ip;
  arp_list[0].usecnt = 1;
#endif
}



#ifndef NOARP
/*
 **************************************************************************
 * 
 * ARP receiver routine
 */
static int arp_recv(char *buf, int bufsize, int typeval, char *addr)
{
#define ahp ((struct arphdr *)buf)

  /*
   * Do some checks on the received packet. It must be an ARP packet,
   * with an ethernet hardware address, and for the IP protocol. Other
   * packets just get discarded. Also discard all packets which have the
   * same sender hardware address as our ethernet card, or which use
   * a broadcast address as sender.
   */
  if (bufsize < sizeof(struct arphdr) || typeval != htons(ETH_P_ARP) ||
      ahp->ar_hrd != htons(ARPHRD_ETHER) || ahp->ar_hln != ETH_ALEN ||
      ahp->ar_pro != htons(ETH_P_IP) || ahp->ar_pln != IP_ALEN ||
      !memcmp(ahp->ar_sha, myhwaddr, ETH_ALEN) ||
      !memcmp(ahp->ar_sha, bcasthw, ETH_ALEN) ||
       memcmp(ahp->ar_sha, addr, ETH_ALEN))
	return(PKT_DISCARD);

  /* If the packet is for us, save the sender's address */
  if (!memcmp(ahp->ar_tha, myhwaddr, ETH_ALEN))
	addcache(ahp->ar_sha, _ntohl(ahp->ar_sip));

  /* If the packet is a request, send our own hardware address back */
  if (ahp->ar_op == htons(ARPOP_REQUEST) &&
      ahp->ar_tip == htonl(myipaddr)) {
		ahp->ar_op  = htons(ARPOP_REPLY);
		ahp->ar_tip = ahp->ar_sip;
		ahp->ar_sip = htonl(myipaddr);
		memcpy(ahp->ar_tha, ahp->ar_sha, ETH_ALEN);
		memcpy(ahp->ar_sha, myhwaddr, ETH_ALEN);
		write_packet((char *)ahp, sizeof(struct arphdr), typeval,
							(char *)ahp->ar_tha);
  }
  return(PKT_DISCARD);

#undef ahp
}
#endif



/*
 **************************************************************************
 * 
 * Resolve IP address and return pointer to hardware address.
 */
unsigned char *resolve(t_ipaddr ip)
{
#define ahp ((struct arphdr *)arpbuf)

  int i;

  /* Broadcast addresses are always known */
  if (ip == IP_BROADCAST || ip == mybcastip)
	return(bcasthw);

  /* Check for own IP address */
  if (ip == myipaddr || ip == IP_LOCALHOST)
	return(myhwaddr);

  /* Check for invalid IP addresses */
  if (ip == IP_ANY)
	return(NULL);

  /* All other addresses have to be on our local subnet */
  if ((ip ^ myipaddr) & mynetmask)
	return(NULL);

  /* Check if we got the address in the cache already */
  if ((i = findip(ip)) >= 0)
	return(arp_list[i].hwaddr);

#ifndef NOARP
  /* No, we have to broadcast a request */
  ahp->ar_hrd = htons(ARPHRD_ETHER);
  ahp->ar_pro = htons(ETH_P_IP);
  ahp->ar_hln = ETH_ALEN;
  ahp->ar_pln = IP_ALEN;
  ahp->ar_op  = htons(ARPOP_REQUEST);
  ahp->ar_sip = htonl(myipaddr);
  ahp->ar_tip = htonl(ip);
  memcpy(ahp->ar_sha, myhwaddr, ETH_ALEN);
  memset(ahp->ar_tha, 0, ETH_ALEN);
  if (write_packet((char *)ahp, sizeof(struct arphdr), htons(ETH_P_ARP),
							(char *)bcasthw) < 0)
	return(NULL);

  /* Wait for someone to answer the broadcast request */
  set_timeout(ARP_TIMEOUT);
  while (!chk_timeout() && (i = findip(ip)) < 0) ;
  if (i >= 0)
	return(arp_list[i].hwaddr);

#endif

  return(NULL);

#undef ahp
}



/*
 **************************************************************************
 * 
 * Network intialization:
 */
int init_arp(void)
{
  /* Set name of module for error messages */
  module_name = "arp";

#ifndef NOARP
  /* Register ARP packet type */
  if (reg_type(htons(ETH_P_ARP), arp_recv) != 0)
	return(FALSE);

  /* Allocate memory for the ARP send buffer */
  if ((arpbuf = malloc(sizeof(struct arphdr))) == NULL)
	return(FALSE);
#endif

  /* Allocate memory for ARP cache */
  if ((arp_list = malloc(ARPNUM * sizeof(struct arp_cache))) == NULL)
	return(FALSE);
  memset(arp_list, 0, ARPNUM * sizeof(struct arp_cache));

  return(TRUE);
}

