/*
 **************************************************************************
 *
 * Boot-ROM-Code to load an operating system across a TCP/IP network.
 *
 * Module:  ip.c
 * Purpose: Internet Protocol handling, includes ICMP handling
 * Entries: init_ip, ip_reg_proto, ip_get_err, ip_send
 *
 **************************************************************************
 *
 * Copyright (C) 1995 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "../headers/packet.h"
#include "../headers/net.h"
#include "../headers/romlib.h"
#include "../headers/general.h"
#include "./netpriv.h"
#include "./icmp.h"
#include "./arp.h"
#include "./ip.h"
#include "./route.h"


/*
 **************************************************************************
 * 
 * Definition of data structure used for the higher protocol without
 * ICMP. Note that we only support one protocol at a time (which is
 * likely to be UDP). This should be sufficient for a boot rom.
 */
struct s_proto {
	unsigned short  proto;		/* protocol ID			*/
	unsigned short  perror;		/* last received error		*/
	proto_recv     *phandler;	/* receive handler		*/
};



/*
 **************************************************************************
 * 
 * Global variables:
 */
static struct s_proto pdef;		/* protocol definition		*/
static char *sendbuf;			/* send buffer			*/
static int pktmtu;			/* MTU as given by packet drvr	*/
static int packetid = 1;		/* packet ID counter		*/



/*
 **************************************************************************
 * 
 * Register a new protocol
 */
int ip_reg_proto(unsigned short proto, proto_recv *handler)
{
  pdef.proto = proto;
  pdef.phandler = handler;
  pdef.perror = 0;
  return(pktmtu);
}



/*
 **************************************************************************
 * 
 * Return code for last error received for a specific protocol via ICMP.
 */
int ip_get_err(void)
{
  int err;

  err = pdef.perror;
  pdef.perror = 0;
  return(err);
}



/*
 **************************************************************************
 * 
 * Send an IP packet.
 */
static void ip_send_pkt(char *buf, int size, int ttl, t_ipaddr dest, char *addr,
							unsigned short proto)
{
#define ipp ((struct iphdr *)sendbuf)

  /*
   * Don't send packets to myself, and also don't send broadcast ICMP packets.
   */
  if (!memcmp(addr, myhwaddr, ETH_ALEN) ||
      (proto == IPPROTO_ICMP && !memcmp(addr, bcasthw, ETH_ALEN)))
	return;

  /* Prepare IP header */
  ipp->ip_ihl_version = ((IP_VERSION << 4) | ((IP_MIN_HSIZE/4) & 0x0f)) & 0xff;
  ipp->ip_tos         = IP_STD_TOS;
  ipp->ip_tot_len     = htons(IP_MIN_HSIZE + size);
  ipp->ip_id          = htons(packetid++);
  ipp->ip_frag_off    = 0;
  ipp->ip_ttl         = ttl;
  ipp->ip_protocol    = proto;
  ipp->ip_check       = 0;
  ipp->ip_daddr       = dest;
  ipp->ip_saddr       = htonl(myipaddr);
  ipp->ip_check       = ipchksum((unsigned char *)sendbuf, IP_MIN_HSIZE);

  /* Copy the buffer contents into the output buffer and send the packet */
  if ((sendbuf + IP_MIN_HSIZE) != buf)
	memcpy(sendbuf + IP_MIN_HSIZE, buf, size);
  (void)write_packet(sendbuf, IP_MIN_HSIZE + size, htons(ETH_P_IP), addr);

#undef ipp
}



#ifndef NOICMP
/*
 **************************************************************************
 * 
 * ICMP receiver routine
 */
static void icmp_recv(char *buf, char *addr)
{
#define ipp ((struct iphdr *)buf)

  struct icmphdr *icmpp;
  t_ipaddr gw;
  t_ipaddr dest;
  unsigned short proto;
  int iphdrlen;
  int icmplen;
  int i;

  /*
   * At this point we can assume a correct IP packet, so no further
   * checks are required except checksum.
   */
  ipp = (struct iphdr *)buf;
  iphdrlen = (ipp->ip_ihl_version & 0x0f) * 4;
  icmplen = ntohs(ipp->ip_tot_len) - iphdrlen;
  icmpp = (struct icmphdr *) &buf[iphdrlen];
  if (ipchksum((unsigned char *)icmpp, icmplen) != 0)
	return;

  /*
   * Handle kind of ICMP packet. Note that most of the packets get
   * discarded, because we only handle a limited number of ICMP
   * message types.
   */
  switch(icmpp->icmp_type) {
	case ICMP_SRC_QUENCH:
	case ICMP_DEST_UNRCH:
	case ICMP_TTL_EXCEED:
	case ICMP_PARAM_PROB:
		proto = ((struct iphdr *) &(icmpp->icmp_val2))->ip_protocol;
		if (proto != pdef.proto) break;
		pdef.perror = (icmpp->icmp_type << 8) & 0xff00;
		pdef.perror |= icmpp->icmp_code & 0x00ff;
		break;
#ifndef NOROUTE
	case ICMP_REDIRECT:
		dest = ntohl(((struct iphdr *) &(icmpp->icmp_val2))->ip_daddr);
		gw   = ntohl((t_ipaddr)icmpp->icmp_val1);
		/* Only add routes for hosts, not for networks */
		if (icmpp->icmp_code == ICMP_RD_HOST)
			rt_add(dest, gw, IP_BROADCAST);
		break;
#endif
	case ICMP_ECHO_REQ:
		icmpp->icmp_type = ICMP_ECHO_REP;
		icmpp->icmp_check = 0;
		icmpp->icmp_check = ipchksum((unsigned char *)icmpp, icmplen);
		ip_send_pkt((char *)icmpp, icmplen, ICMP_TTL, ipp->ip_saddr,
							addr, IPPROTO_ICMP);
		break;
  }

#undef ipp
}
#endif



#ifndef NOICMP
/*
 **************************************************************************
 * 
 * Send an ICMP reply to an incoming IP packet
 */
static void icmp_reply(char *buf, char *addr, int type, int code, ...)
{
#define ipp ((struct iphdr *)buf)

  struct icmphdr *icmpp;
  char *cp;
  void *args;
  int icmplen;

  args = (...);

  /*
   * Only reply to packets if they are not ICMP packets, and they are
   * definitely for me.
   */
  if (ntohl(ipp->ip_daddr) != myipaddr || ipp->ip_protocol == IPPROTO_ICMP)
	return;

  /*
   * Copy the original internet header at the end of the ICMP header. This
   * is necessary for all message types. It is not as trivial as it might
   * sound, because the copy routine overwrites from bottom to top. That
   * means we have to copy the relevant buffer area behind the destination
   * area, and then back to it's final destination. This works assuming
   * that the maximum IP header size does not exceed 64 bytes, and the
   * send buffer has a size of greater than 256 bytes. No checks are
   * made here if this is really true!
   */
  icmplen = (ipp->ip_ihl_version & 0x0f) * 4 + ICMP_DATA_SIZE;
  icmpp = (struct icmphdr *)(buf + IP_MIN_HSIZE);
  cp = buf + IP_SEND_SIZE - icmplen - 1;
  memcpy(cp, buf, icmplen);
  memcpy((char *) &(icmpp->icmp_val2), cp, icmplen);
  icmplen += (int) &(icmpp->icmp_val2) - (int)icmpp;

  /*
   * Now process the ICMP message type and setup the ICMP header. The
   * arguments to this function depend on this message type.
   */
  icmpp->icmp_type  = type;
  icmpp->icmp_code  = code;
  switch(type) {
	case ICMP_PARAM_PROB:
		(unsigned char)(icmpp->icmp_val1) = (*((int *)args)++) & 0xff;
		break;
	case ICMP_REDIRECT:
		(t_ipaddr)(icmpp->icmp_val1) = htonl(*((t_ipaddr *)args)++);
		break;
	default:
		icmpp->icmp_val1 = 0;
		break;
  }
  icmpp->icmp_check = 0;
  icmpp->icmp_check = ipchksum((unsigned char *)icmpp, icmplen);

  /* Finally send the packet */
  ip_send_pkt((char *)icmpp, icmplen, ICMP_TTL, ipp->ip_saddr, addr,
								IPPROTO_ICMP);

#undef ipp
}
#endif



/*
 **************************************************************************
 * 
 * IP receiver routine
 */
static int ip_recv(char *buf, int bufsize, int typeval, char *addr, int reuse)
{
#define ipp ((struct iphdr *)buf)

  int i, size;
  int hdrlen;
  t_ipaddr dadr;

  /*
   * Do some checks on the received packet. It must be an IP packet
   * with a valid version number and header size. Also the checksum
   * of the header must be correct, and the size value must not be
   * greater than the actual packet size.
   */
  size = ntohs(ipp->ip_tot_len);
  hdrlen = (ipp->ip_ihl_version & 0x0f) * 4;
  if (bufsize <= sizeof(struct iphdr) || typeval != htons(ETH_P_IP) ||
      hdrlen < IP_MIN_HSIZE || ipchksum((unsigned char *)ipp, hdrlen) != 0 ||
      (ipp->ip_ihl_version & 0xf0) != (IP_VERSION << 4) ||
      size > bufsize || size > pktmtu)
	return(PKT_DISCARD);

  /*
   * The destination address of the packet must be valid. It must either
   * be 0.0.0.0 (which means "any"), 255.255.255.255 for broadcast packets,
   * or our own IP number.
   */
  dadr = ntohl(ipp->ip_daddr);
  if (dadr != IP_ANY && dadr != IP_BROADCAST && dadr != myipaddr &&
      myipaddr != IP_ANY)
	return(PKT_DISCARD);

  /*
   * We don't accept packets, which are fragmented, because fragmentation
   * should not be necessary on an ethernet. The Low Level ethernet frames
   * are large enough to hold more than a minimum IP packet. This check also
   * ensures, that ICMP messages lateron will only be sent in response to
   * the first IP packet, and not to successive fragments.
   */
  if ((ntohs(ipp->ip_frag_off) & 0x3fff) != 0) {
#ifndef NOICMP
	/* Only respond with ICMP reply to first fragment */
	if ((ntohs(ipp->ip_frag_off) & 0x3fff) == 0x2000)
		icmp_reply(buf, addr, ICMP_PARAM_PROB, 0,
					(int) &(ipp->ip_frag_off) - (int)ipp);
#endif
	return(PKT_DISCARD);
  }

  /*
   * We don't accept packets with options.
   */
  if (hdrlen > IP_MIN_HSIZE) {
#ifndef NOICMP
	icmp_reply(buf, addr, ICMP_PARAM_PROB, 0, IP_MIN_HSIZE);
#endif
	return(PKT_DISCARD);
  }

  /*
   * Handle ICMP packets
   */
#ifndef NOICMP
  if (ipp->ip_protocol == IPPROTO_ICMP) {
	icmp_recv(buf, addr);
	return(PKT_DISCARD);
  }
#endif

  /*
   * OK. Now we know, that the IP packet is valid and indeed meant for us.
   * Check if it's for a supported protocol.
   */
  if (pdef.proto != ipp->ip_protocol || pdef.phandler == NULL) {
#ifndef NOICMP
	icmp_reply(buf, addr, ICMP_DEST_UNRCH, ICMP_PROT_UNRCH);
#endif
	return(PKT_DISCARD);
  }

#ifdef NOARP
  /*
   * If we don't have implemented the ARP protocol, there must be another
   * way to associate an IP address to a hardware address. The way this
   * bootrom normally works is by sending out a broadcast BOOTP request
   * and then waiting for an answer. The first packet received has the
   * IP and hardware addresses of the server, so use that information
   * to update the ARP cache.
   */
  addcache(addr, ntohl(ipp->ip_saddr));
#endif

  /*
   * Send the full packet (including IP header) up one layer. If the
   * packet should be kept, notify sender with a Source Quench ICMP
   * message. Note that the upper layer routine is not allowed to
   * change anything within the data packet.
   */
  if (pdef.phandler(buf, size) == IP_KEEP) {
#ifndef NOICMP
	if (!reuse) {
		memcpy(sendbuf, buf, size);
		icmp_reply(sendbuf, addr, ICMP_SRC_QUENCH, 0);
	}
#endif
	return(PKT_KEEP);
  }

  return(PKT_DISCARD);

#undef ipp
}



/*
 **************************************************************************
 *
 * Send an IP packet.
 */
int ip_send(char *buf, int size, t_ipaddr dest)
{
  char *addr;

  /* Find a routing to the destination host */
  if (dest == myipaddr)
	return(size);
  if ((addr = (char *)resolve(find_rt(dest))) == NULL)
	return(0);

  /* Do not send more bytes than fit into a standard IP frame */
  if (size > (IP_SEND_SIZE - IP_MIN_HSIZE))
	size = IP_SEND_SIZE - IP_MIN_HSIZE;

#ifdef DEBUG
  printf("IP: sending %d bytes to %x:%x:%x:%x:%x:%x\n",
	size, addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);
#endif

  /* And finally send the packet */
  ip_send_pkt(buf, size, IP_TTL, htonl(dest), addr, pdef.proto);

  return(size);
}



/*
 **************************************************************************
 * 
 * Network intialization:
 */
int init_ip(void)
{
  /* Set module name for error handling */
  module_name = "ip";

  /* Register IP packet type and set the MTU for the interface */
  if (reg_type(htons(ETH_P_IP), ip_recv) != 0)
	return(FALSE);
  if ((pktmtu = get_mtu()) < IP_RECV_SIZE)
	return(FALSE);

  /* Allocate memory for the send buffer and initialize the receive handlers */
  if ((sendbuf = malloc(IP_SEND_SIZE)) == NULL)
	return(FALSE);
  pdef.phandler = NULL;
  pdef.proto = 0;
  pdef.perror = 0;

  return(TRUE);
}

