/*
 **************************************************************************
 *
 * Boot-ROM-Code to load an operating system across a TCP/IP network.
 *
 * Module:  route.c
 * Purpose: Handle routing information.
 * Entries: rt_add, find_rt
 *
 **************************************************************************
 *
 * Copyright (C) 1995 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "../headers/packet.h"
#include "../headers/net.h"
#include "../headers/romlib.h"
#include "../headers/general.h"
#include "./netpriv.h"
#include "./route.h"



/*
 **************************************************************************
 * 
 * Global variables
 */
static struct rt_table *rt_list = NULL;		/* routing table	*/
static int table_count = 1;			/* table timeout cnt	*/



#ifndef NOROUTE
/*
 **************************************************************************
 * 
 * Find an IP address in the routing table. Return -1 if not found.
 */
static int findip(t_ipaddr ip)
{
  int i, defrt;

  defrt = -1;
  for (i = 0; i < RTNUM; i++)
	if (rt_list[i].usecnt > 0) {
		if (rt_list[i].ipaddr == ip)
			return(i);
		if (rt_list[i].ipaddr == IP_DEFAULT && defrt < 0) {
			defrt = i;
			continue;
		}
		if (rt_list[i].netmask == IP_BROADCAST) continue;
		if (!((rt_list[i].ipaddr ^ ip) & rt_list[i].netmask))
			return(i);
	}

  return(defrt);
}
#endif



/*
 **************************************************************************
 * 
 * Find an IP address in the routing table, and return that address to
 * which to direct the ethernet packets (e.g. either destination host
 * gateway). Returns IP_ANY if not found.
 */
t_ipaddr find_rt(t_ipaddr dest)
{
  int i;

  /* Handle special addresses first. */
  if (dest == IP_BROADCAST || dest == mybcastip)
	return(mybcastip);
  if (dest == IP_ANY || dest == myipaddr || dest == IP_LOCALHOST)
	return(IP_ANY);

  /* If the destination is on our subnet, send the packets directly */
  if (!((dest ^ myipaddr) & mynetmask))
	return(dest);

#ifndef NOROUTE
  /* Now search for the address in the routing table */
  if ((i = findip(dest)) < 0)
	return(IP_ANY);

  return(rt_list[i].gwaddr);
#else
  return(IP_ANY);
#endif
}



/*
 **************************************************************************
 * 
 * Add a new entry to the routing table.
 */
void rt_add(t_ipaddr dest, t_ipaddr gw, t_ipaddr mask)
{
#ifndef NOROUTE
  int i, oldest, oldval;

  /*
   * If the destination is on our own local network, we don't need
   * a gateway, and network routes are not allowed. If the destination
   * is not on our local network, then a gateway is required, which
   * has to be reachable.
   */
  if (dest == mybcastip || dest == IP_BROADCAST || dest == IP_LOCALHOST) return;

  if (dest == IP_DEFAULT || (dest ^ myipaddr) & mynetmask) {
	if (gw == IP_ANY || gw == mybcastip ||
	    gw == IP_BROADCAST || gw == IP_LOCALHOST) return;
	if ((gw ^ myipaddr) & mynetmask) {
		/* Find a host route to the gateway if not on our subnet */
		if ((i = findip(gw)) < 0) return;
		if (rt_list[i].netmask != IP_BROADCAST) return;
		if (rt_list[i].gwaddr != IP_ANY)
			gw = rt_list[i].gwaddr;
	}
	/* Ensure that the gateway really is reachable on our subnet */
	if ((gw ^ myipaddr) & mynetmask) return;
  } else
	/* Hosts on our own network don't need a special route */
	return;

  /*
   * If the netmask is not given, try to determine one from the class
   * of the destination address.
   */
  if (mask == IP_ANY) {
	if (dest == IP_DEFAULT)
		mask = IP_ANY;
	else if (IN_CLASS_A(dest))
		mask = IP_CLASS_A;
	else if (IN_CLASS_B(dest))
		mask = IP_CLASS_B;
	else if (IN_CLASS_C(dest))
		mask = IP_CLASS_C;
  }

  /*
   * Search for the oldest entry in the table. Also delete all
   * old entries with the destination IP address.
   */
  oldest = 0;
  oldval = table_count;
  for (i = 1; i < RTNUM; i++) {
	if (rt_list[i].ipaddr == dest)
		rt_list[i].usecnt = 0;
	if (rt_list[i].usecnt < oldval) {
		oldest = i;
		oldval = rt_list[i].usecnt;
	}
  }

  /* Then put the new entry into the table */
  rt_list[oldest].ipaddr  = dest;
  rt_list[oldest].gwaddr  = gw;
  rt_list[oldest].netmask = mask;
  rt_list[oldest].usecnt  = table_count++;
#endif
}



/*
 **************************************************************************
 * 
 * Network intialization:
 */
int init_route(void)
{
  /* Set name of module for error messages */
  module_name = "route";

#ifndef NOROUTE
  /* Allocate memory for routing table */
  if ((rt_list = malloc(RTNUM * sizeof(struct rt_table))) == NULL)
	return(FALSE);
  memset(rt_list, 0, RTNUM * sizeof(struct rt_table));
#endif

  return(TRUE);
}

