/*
 **************************************************************************
 *
 * Boot-ROM-Code to load an operating system across a TCP/IP network.
 *
 * Module:  udp.c
 * Purpose: UDP handling
 * Entries: init_udp, udp_open, udp_close, udp_read, udp_write
 *
 **************************************************************************
 *
 * Copyright (C) 1995 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "../headers/packet.h"
#include "../headers/net.h"
#include "../headers/romlib.h"
#include "../headers/general.h"
#include "./netpriv.h"
#include "./ip.h"
#include "./icmp.h"
#include "./udp.h"


/*
 **************************************************************************
 * 
 * Structure used for holding UDP socket information
 */
struct udp_sock {
	unsigned short source;		/* Source port number		*/
	unsigned short dest;		/* Destination port number	*/
	t_ipaddr       daddr;		/* Destination IP address	*/
	char          *buf;		/* pointer to receive buffer	*/
	char          *readp;		/* read pointer			*/
	int            readlen;		/* length of contents in buffer	*/
};


/*
 **************************************************************************
 * 
 * Global variables:
 */
static struct udp_sock usock;		/* socket			*/
static int recv_buf_size;		/* size of receive buffer	*/
static char *writebuf;			/* output buffer		*/
#ifndef NOICMP
static int udp_errors = 0;		/* counter for UDP errors	*/
static int src_quench = 0;		/* source quench counter	*/
static int src_count = 0;		/* counter for reducing quench	*/
static int src_timer = 0;		/* timer value for each quench	*/
#endif



/*
 **************************************************************************
 * 
 * Open a new UDP socket.
 */
void udp_open(t_ipaddr daddr, int source, int dest)
{
  usock.source = source;
  usock.dest = dest;
  usock.daddr = daddr;
}



/*
 **************************************************************************
 * 
 * UDP receiver routine. This gets called by the IP layer when a packet
 * arrives. It returns IP_KEEP if the packet cannot get discarded.
 */
static int udp_recv(char *buf, int size)
{
#define ipp ((struct iphdr *)buf)

  int i;
  int iphdrlen;
  struct udphdr *udpp;

  /* First check if the checksum is correct */
  iphdrlen = (ipp->ip_ihl_version & 0x0f) * 4;
  udpp = (struct udphdr *)(buf + iphdrlen);
  if (udpp->udp_check != 0) {
	/*
	 * If the checksum in the packet is 0, no checks have to be done. The
	 * 0 checksum value is represented by 0xffff. If we change the checksum
	 * value to 0 here, we can prevent checking again in case the packet
	 * has to be kept.
	 */
	i = udpp->udp_check;
	udpp->udp_check = 0;
	if (utchksum(ntohl(ipp->ip_saddr), ntohl(ipp->ip_daddr), IPPROTO_UDP,
					size - iphdrlen, (char *)udpp) != i)
		return(IP_DISCARD);
  }

  /*
   * Check if we have a suitable socket for this packet. Also check if
   * the length of the packet is valid. Otherwise discard the packet.
   */
  if (usock.source != ntohs(udpp->udp_dest) ||
      (sizeof(struct udphdr) + iphdrlen) > size ||
      (ntohs(udpp->udp_len) + iphdrlen) > size)
	return(IP_DISCARD);

  /* Check if the buffer is free */
  if (usock.readlen > 0)
	return(IP_KEEP);

  /* Update the sockets destination port number */
  usock.dest = ntohs(udpp->udp_source);

  /* Now actually get the packet from the IP layer */
  usock.readlen = ntohs(udpp->udp_len) - sizeof(struct udphdr);
  usock.readp = usock.buf + sizeof(struct udphdr);
  memcpy(usock.buf, (char *)udpp, size - iphdrlen);

  return(IP_DISCARD);

#undef ipp
}



#ifndef NOICMP
/*
 **************************************************************************
 * 
 * Handle UDP protocol errors. We just count the errors, but don't do
 * anything else yet.
 */
static void check_err(void)
{
  int i;

  if ((i = ip_get_err()) != 0) {
	if (((i >> 8) & 0x00ff) == ICMP_SRC_QUENCH) {
		src_quench += (4 * src_timer);
		src_count = 0;
	} else {
		udp_errors++;
	}
  }
}
#endif



/*
 **************************************************************************
 * 
 * Read some bytes from a UDP socket
 */
int udp_read(char *buf, int len, int timeout)
{
  int copied, i;

  /* Length must be >0 */
  if (len == 0)
	return(0);

  /* Wait until we get something */
  copied = 0;
  if (timeout > 0) set_timeout(timeout);
  while (timeout == 0 || (!chk_timeout() && len > 0)) {
#ifndef NOICMP
	check_err();
#endif
  	i = len > usock.readlen ? usock.readlen : len;
	if (i > 0) {
		memcpy(buf, usock.readp, i);
		usock.readp += i;
		usock.readlen -= i;
		copied += i;
		buf += i;
		len -= i;
	}
	if (timeout == 0) break;
  }
  return(copied);
}



/*
 **************************************************************************
 * 
 * Write some data to a UDP socket
 */
int udp_write(char *buf, int len)
{
#define udpp ((struct udphdr *)writebuf)

  int copied, i, quench;

  /* Length must be >0 */
  if (len == 0)
	return(0);

  /* Prepare UDP datagram and send it out */
  udpp->udp_source = htons(usock.source);
  udpp->udp_dest = htons(usock.dest);

  copied = 0;
  while (len > 0) {
#ifndef NOICMP
	check_err();
#endif
	i = len > UDP_SEND_SIZE ? UDP_SEND_SIZE : len;
	memcpy(((char *)udpp) + sizeof(struct udphdr), buf, i);
	copied += i;
	buf += i;
	len -= i;
	i += sizeof(struct udphdr);
	udpp->udp_len = htons(i);
	udpp->udp_check = 0;
	udpp->udp_check = utchksum(myipaddr, usock.daddr,
						IPPROTO_UDP, i, (char *)udpp);
#ifndef NOICMP
	/* Care for source quench requests by cutting down the output rate */
	if (src_quench > 0) {
		if (src_count++ > 10) {		/* Reduce quench after 10 pkt */
			src_quench -= src_timer;
			src_count = 0;
		}
		for (quench = src_quench; quench > 0; quench--) ;
	}
#endif
#ifdef DEBUG
	printf("UDP: sending from %ld.%d to %ld.%d\n",
			myipaddr, usock.source, usock.daddr, usock.dest);
#endif
	/* Now send the datagram */
	if (ip_send((char *)udpp, i, usock.daddr) != i)
		return(copied - i);
  }
  return(copied);

#undef udpp
}



/*
 **************************************************************************
 * 
 * Network intialization:
 */
int init_udp(void)
{
  char *cp;
  int i;

  /* Set module name for error handling */
  module_name = "udp";

  /* Register the UDP protocol with the IP layer */
  if ((recv_buf_size = ip_reg_proto(IPPROTO_UDP, udp_recv)) == 0)
	return(FALSE);

  /* Allocate memory for the receive and send buffers */
  if ((cp = malloc(IP_SEND_SIZE + recv_buf_size)) == NULL)
	return(FALSE);
  usock.buf = cp;
  usock.source = 0;		/* indicates free slot		*/
  usock.readlen = 0;		/* indicates free buffer	*/
  usock.readp = cp;
  writebuf = cp + recv_buf_size;

#ifndef NOICMP
  /* Determine number of simple loops for each source quench request */
  set_timeout(2);
  for (i = 0; !chk_timeout(); i++) ;
  src_timer = i < 180 ? 1 : (i / 180);	/* should give approx. 2 ms	*/
#endif

  return(TRUE);
}

