;**************************************************************************
;*
;* Boot-ROM-Code to load an operating system across a TCP/IP network.
;*
;* Module:  loadpkt.asm
;* Purpose: Load a packet driver, optionally with decompression
;* Entries: loadpkt
;*
;**************************************************************************
;*
;* Copyright (C) 1995 Gero Kuhlmann <gero@gkminix.han.de>
;*
;*  This program is free software; you can redistribute it and/or modify
;*  it under the terms of the GNU General Public License as published by
;*  the Free Software Foundation; either version 2 of the License, or
;*  any later version.
;*
;*  This program is distributed in the hope that it will be useful,
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;*  GNU General Public License for more details.
;*
;*  You should have received a copy of the GNU General Public License
;*  along with this program; if not, write to the Free Software
;*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;*


;
;**************************************************************************
;
; Include assembler macros:
;
include ..\headers\asm\layout.inc
include ..\headers\asm\macros.inc
include ..\headers\asm\memory.inc
include ..\headers\asm\minidos.inc
include .\pktpriv.inc


;
;**************************************************************************
;
; Start code segment.
;
text_start

	public	loadpkt			; define entry points


; Magic cookie to identify COM programs

cookie	db	COM_MAGIC

;
;**************************************************************************
;
; Load a packet driver into that memory space which is intended for
; holding the booted operating system. It's just a temporary space,
; from where the DOS simulator can get the packet driver code. While
; copying the packet driver from ROM into RAM, it will get decompressed.
; Input:  BX:SI  -  pointer to compressed packet driver
; Output: ES:DI  -  pointer to decompressed packet driver
; Registers changed: AX, BX, CX, DX, SI, DI, ES
;
loadpkt		proc	near

	cld
	push	ds
	mov	ax,LOWMEM		; the boot rom startup code already
	mov	es,ax			; made sure, that there really is RAM
	xor	di,di			; at our destination
	mov	ax,cs			; first copy the magic cookie at
	mov	ds,ax			; the destination memory space
	mov	ax,si
	mov	si,offset cgroup:cookie
	mov	cx,COM_MAGIC_LEN
	rep	movsb

	mov	ds,bx
	mov	si,ax
	mov	di,COM_CODE_OFS
	lodsw				; load magic ID of source code
	cmp	ax,COMPR_MAGIC
	mov	ax,0FFFFh		; check if it's correct
	jne	loadp9
	lodsw				; load compressed image's size
	mov	cx,ax
	lodsb				; load compression method ID
	cmp	al,COMPR_NONE		; is it compressed at all?
	je	loadp1
	cmp	al,COMPR_STD		; standard compression method
	je	loadp2
	mov	ax,0FFFFh		; oops, don't know compression
	jmp	short loadp9		; method

loadp1:	mov	dx,cx			; packet driver is uncompressed
	rep	movsb			; just move it to it's destination
	jmp	short loadp3

loadp2:	call	loadstd			; standard compression method
loadp3:	add	dx,0Fh			; convert number of bytes at
	mov	cl,4			; destination into paragraphs
	shr	dx,cl
	mov	es:[COM_SIZE_OFS],dx	; save size of packet driver image
	xor	di,di
	xor	ax,ax			; return without error

loadp9:	pop	ds
	ret

loadpkt		endp


;
;**************************************************************************
;
; Decompress the packet driver code using standard compression method.
; The compression works by searching for identical byte sequences. The
; compressed image then consists of two types of blocks: literal blocks,
; and blocks with a reversed pointer to a byte sequence which has already
; been decoded. Preceeding each block are one or two bytes which are
; encoded as follows:
;
;	literal block  -  0000llll
;	reverse block  -  llllpppp pppppppp
;
; llll means the length of the block minus 1
; p..p means the backwards byte distance in the destination area to the
;      identical looking byte sequence
;
; Of course this compression scheme is not very high sophisticated (compression
; rate with programs usually around 30%), but it's very simple and doesn't
; need a long decompression routine. The packet drivers are usually so small
; that compressing them with a higher sophisticated but significantly larger
; decompression routine would not gain any real space saving in the ROM.
;
; Input:  DS:SI  -  Compressed source code
;         ES:DI  -  Destination memory area
;         CX     -  Size of compressed source code
; Output: DX     -  Size of decompressed destination code
; Registers changed: AX, BX, CX, DX, Si, DI
;
loadstd		proc	near

	xor	dx,dx		; DX is destination byte counter
	jcxz	loads9		; nothing to read?
loads1:	lodsb			; load identifying byte
	dec	cx
	jz	loads9		; at the end of the input stream?
	mov	ah,al
	test	al,0F0h		; check if it shows a literal byte block
	jnz	loads3

	inc	ah
loads2:	movsb			; copy literal bytes into destination
	inc	dx
	dec	cx		; at the end of the input stream? this
	jz	loads9		; shouldn't happen here!
	dec	ah
	jnz	loads2		; loop over all bytes in the literal block
	jmp	loads1		; continue with next ID byte

loads3:	shr	ah,1		; compute number of bytes in reverse block
	shr	ah,1
	shr	ah,1		; note: unfortunately we cannot use CL here
	shr	ah,1		; for shifting
	inc	ah
	and	al,0Fh		; isolate offset
	mov	bh,al
	lodsb			; get next byte of offset value
	dec	cx
	jz	loads9		; shouldn't happen here!
	mov	bl,al
	neg	bx
	add	bx,di		; subtract offset from current destination
	dec	bx		; pointer
loads4:	mov	al,es:[bx]	; copy reverse reference block
	stosb
	inc	bx
	inc	dx		; update destination byte counter
	dec	ah
	jnz	loads4
	jmp	loads1		; continue with next ID byte

loads9:	ret

loadstd		endp


;
;**************************************************************************
;
text_end

	end

